from . import sedes  # noqa: F401
from .codec import (  # noqa: F401
    encode,
    decode,
    infer_sedes,
)
from .exceptions import (  # noqa: F401
    RLPException,
    EncodingError,
    DecodingError,
    SerializationError,
    DeserializationError,
)
from .lazy import decode_lazy, peek, LazyList  # noqa: F401
from .sedes import Serializable  # noqa: F401



"""
Module for sedes objects that use lists as serialization format.
"""
from collections import Sequence

from eth_utils import (
    to_list,
    to_tuple,
)

from rlp_cython.exceptions import (
    SerializationError,
    ListSerializationError,
    DeserializationError,
    ListDeserializationError,
    RLPException,
)

from .binary import (
    Binary as BinaryClass,
)


def is_sedes(obj):
    """Check if `obj` is a sedes object.

    A sedes object is characterized by having the methods `serialize(obj)` and
    `deserialize(serial)`.
    """
    return hasattr(obj, 'serialize') and hasattr(obj, 'deserialize')


def is_sequence(obj):
    """Check if `obj` is a sequence, but not a string or bytes."""
    return isinstance(obj, Sequence) and not (
        isinstance(obj, str) or BinaryClass.is_valid_type(obj))


class List(list):

    """A sedes for lists, implemented as a list of other sedes objects.

    :param strict: If true (de)serializing lists that have a length not
                   matching the sedes length will result in an error. If false
                   (de)serialization will stop as soon as either one of the
                   lists runs out of elements.

    :param has_serializable_children: Set to true if any children in the list are serializable.
                                      Otherwise, it won't try to serialize any children.
    """

    def __init__(self, elements=None, strict=True, has_serializable_children = True):
        super(List, self).__init__()
        self.strict = strict
        self.has_serializable_children = has_serializable_children

        # if len(elements) < 2:
        #     raise RLPException("List needs at least 2 elements")

        if elements:
            for e in elements:
                if is_sedes(e):
                    self.append(e)
                elif isinstance(e, Sequence):
                    self.append(List(e))
                else:
                    raise TypeError(
                        'Instances of List must only contain sedes objects or '
                        'nested sequences thereof.'
                    )

    def get_sede_identifier(self):
        inner_list = []
        for sede in self:
            try:
                inner_list.append(sede.get_sede_identifier())
            except AttributeError:
                inner_list.append(0)

        return inner_list


    # @to_list
    # def serialize(self, obj):
    #     if self.validate and not is_sequence(obj):
    #         raise ListSerializationError('Can only serialize sequences', obj)
    #     if self.strict:
    #         if len(self) != len(obj) or len(self) < len(obj):
    #             raise ListSerializationError('List has wrong length', obj)
    #
    #     for index, (element, sedes) in enumerate(zip(obj, self)):
    #         try:
    #             yield sedes.serialize(element)
    #         except SerializationError as e:
    #             raise ListSerializationError(obj=obj, element_exception=e, index=index)
    #
    # @to_tuple
    # def deserialize(self, serial):
    #     if self.validate and not is_sequence(serial):
    #         raise ListDeserializationError('Can only deserialize sequences', serial)
    #
    #     if self.strict and len(serial) != len(self):
    #         raise ListDeserializationError('List has wrong length', serial)
    #
    #     for idx, (sedes, element) in enumerate(zip(self, serial)):
    #         try:
    #             yield sedes.deserialize(element)
    #         except DeserializationError as e:
    #             raise ListDeserializationError(serial=serial, element_exception=e, index=idx)

    @to_list
    def serialize_to_list(self, obj):
        for index, (element, sedes) in enumerate(zip(obj, self)):
            try:
                yield sedes.serialize(element)
            except SerializationError as e:
                raise ListSerializationError(obj=obj, element_exception=e, index=index)

    def serialize(self, obj):
        if self.has_serializable_children:
            return self.serialize_to_list(obj)
        else:
            return obj

    @to_tuple
    def deserialize_to_tuple(self, serial):
        for idx, (sedes, element) in enumerate(zip(self, serial)):
            try:
                yield sedes.deserialize(element)
            except DeserializationError as e:
                raise ListDeserializationError(serial=serial, element_exception=e, index=idx)

    @to_list
    def deserialize_to_list(self, serial):
        for idx, (sedes, element) in enumerate(zip(self, serial)):
            try:
                yield sedes.deserialize(element, to_list = True)
            except DeserializationError as e:
                raise ListDeserializationError(serial=serial, element_exception=e, index=idx)

    def deserialize(self, serial, to_list = False):
        if self.has_serializable_children:
            if to_list:
                return self.deserialize_to_list(serial)
            else:
                return self.deserialize_to_tuple(serial)
        else:
            return serial


#use this when all of the children serialization and deserialzation is handled with the cython back end
def FList(elements):
    return List(elements, has_serializable_children = False)

class CountableList(object):

    """A sedes for lists of arbitrary length.

    :param element_sedes: when (de-)serializing a list, this sedes will be
                          applied to all of its elements
    :param max_length: maximum number of allowed elements, or `None` for no limit
    """

    def __init__(self, element_sedes, has_serializable_children = True):

        self.element_sedes = element_sedes
        self.has_serializable_children = has_serializable_children

    def get_sede_identifier(self):
        try:
            return [self.element_sedes.get_sede_identifier()]
        except AttributeError:
            return [0]




    # @to_list
    # def serialize(self, obj):
    #     if self.validate and not is_sequence(obj):
    #         raise ListSerializationError('Can only serialize sequences', obj)
    #
    #     if self.max_length is not None and len(obj) > self.max_length:
    #         raise ListSerializationError(
    #             'Too many elements ({}, allowed {})'.format(
    #                 len(obj),
    #                 self.max_length,
    #             ),
    #             obj=obj,
    #         )
    #
    #     for index, element in enumerate(obj):
    #         try:
    #             yield self.element_sedes.serialize(element)
    #         except SerializationError as e:
    #             raise ListSerializationError(obj=obj, element_exception=e, index=index)
    #
    # @to_tuple
    # def deserialize(self, serial):
    #     if self.validate and not is_sequence(serial):
    #         raise ListDeserializationError('Can only deserialize sequences', serial=serial)
    #     for index, element in enumerate(serial):
    #         if self.max_length is not None and index >= self.max_length:
    #             raise ListDeserializationError(
    #                 'Too many elements (more than {})'.format(self.max_length),
    #                 serial=serial,
    #             )
    #
    #         try:
    #             yield self.element_sedes.deserialize(element)
    #         except DeserializationError as e:
    #             raise ListDeserializationError(serial=serial, element_exception=e, index=index)

    @to_list
    def serialize_to_list(self, obj):
        for index, element in enumerate(obj):
            try:
                yield self.element_sedes.serialize(element)
            except SerializationError as e:
                raise ListSerializationError(obj=obj, element_exception=e, index=index)

    @to_tuple
    def deserialize_to_tuple(self, serial):
        for index, element in enumerate(serial):
            try:
                yield self.element_sedes.deserialize(element)
            except DeserializationError as e:
                raise ListDeserializationError(serial=serial, element_exception=e, index=index)

    @to_list
    def deserialize_to_list(self, serial):
        for index, element in enumerate(serial):
            try:
                yield self.element_sedes.deserialize(element, to_list = True)
            except DeserializationError as e:
                raise ListDeserializationError(serial=serial, element_exception=e, index=index)


    def serialize(self, obj):
        if self.has_serializable_children:
            return self.serialize_to_list(obj)
        else:
            return obj

    def deserialize(self, serial, to_list = False):
        if self.has_serializable_children:
            if to_list:
                return self.deserialize_to_list(serial)
            else:
                return self.deserialize_to_tuple(serial)
        else:
            return serial

#use this when all of the children serialization and deserialzation is handled with the cython back end
def FCountableList(elements):
    return CountableList(elements, has_serializable_children = False)


import binascii

import pytest

from eth_utils import int_to_big_endian

from rlp_cython import SerializationError
from rlp_cython.sedes import big_endian_int, BigEndianInt
from rlp_cython.utils import ALL_BYTES


valid_data = (
    (256, b'\x01\x00'),
    (1024, b'\x04\x00'),
    (65535, b'\xff\xff'),
)

single_bytes = ((n, ALL_BYTES[n]) for n in range(1, 256))

random_integers = (256, 257, 4839, 849302, 483290432, 483290483290482039482039,
                   48930248348219540325894323584235894327865439258743754893066)
assert random_integers[-1] < 2**256

negative_ints = (-1, -100, -255, -256, -2342423)


def test_neg():
    for n in negative_ints:
        with pytest.raises(SerializationError):
            big_endian_int.serialize(n)


@pytest.mark.parametrize('value', [True, False])
def test_rejects_bool(value):
    with pytest.raises(SerializationError):
        big_endian_int.serialize(value)


def test_serialization():
    for n in random_integers:
        serial = big_endian_int.serialize(n)
        deserialized = big_endian_int.deserialize(serial)
        assert deserialized == n
        if n != 0:
            assert serial[0] != b'\x00'  # is not checked


def test_single_byte():
    for n, s in single_bytes:
        serial = big_endian_int.serialize(n)
        assert serial == s
        deserialized = big_endian_int.deserialize(serial)
        assert deserialized == n


def test_valid_data():
    for n, serial in valid_data:
        serialized = big_endian_int.serialize(n)
        deserialized = big_endian_int.deserialize(serial)
        assert serialized == serial
        assert deserialized == n


def test_fixedlength():
    s = BigEndianInt(4)
    for i in (0, 1, 255, 256, 256**3, 256**4 - 1):
        assert len(s.serialize(i)) == 4
        assert s.deserialize(s.serialize(i)) == i
    for i in (256**4, 256**4 + 1, 256**5, -1, -256, 'asdf'):
        with pytest.raises(SerializationError):
            s.serialize(i)


def packl(lnum):
    """Packs the lnum (which must be convertable to a long) into a
       byte string 0 padded to a multiple of padmultiple bytes in size. 0
       means no padding whatsoever, so that packing 0 result in an empty
       string.  The resulting byte string is the big-endian two's
       complement representation of the passed in long."""

    if lnum == 0:
        return b'\0'
    s = hex(lnum)[2:]
    s = s.rstrip('L')
    if len(s) & 1:
        s = '0' + s
    s = binascii.unhexlify(s)
    return s


try:
    import ctypes
    PyLong_AsByteArray = ctypes.pythonapi._PyLong_AsByteArray
    PyLong_AsByteArray.argtypes = [ctypes.py_object,
                                   ctypes.c_char_p,
                                   ctypes.c_size_t,
                                   ctypes.c_int,
                                   ctypes.c_int]
    import sys
    long_start = sys.maxint + 1

    def packl_ctypes(lnum):
        if lnum < long_start:
            return int_to_big_endian(lnum)
        a = ctypes.create_string_buffer(lnum.bit_length() // 8 + 1)
        PyLong_AsByteArray(lnum, a, len(a), 0, 1)
        return a.raw.lstrip(b'\0')
except AttributeError:
    packl_ctypes = packl


def test_packl():
    for i in range(256):
        v = 2**i - 1
        rc = packl_ctypes(v)
        assert rc == int_to_big_endian(v)
        r = packl(v)
        assert r == int_to_big_endian(v)


def perf():
    import time

    st = time.time()
    for j in range(100000):
        for i in random_integers:
            packl(i)
    print('packl elapsed {}'.format(time.time() - st))

    st = time.time()
    for j in range(100000):
        for i in random_integers:
            packl_ctypes(i)
    print('ctypes elapsed {}'.format(time.time() - st))

    st = time.time()
    for j in range(100000):
        for i in random_integers:
            int_to_big_endian(i)
    print('py elapsed {}'.format(time.time() - st))


if __name__ == '__main__':
    # test_packl()
    perf()



# -*- coding: UTF-8 -*-
import pytest
from rlp_cython import SerializationError
from rlp_cython.sedes import Binary


@pytest.mark.parametrize(
    'value,expected',
    (
        (b'', b''),
        (b'asdf', b'asdf'),
        (b'\x00' * 20, b'\x00' * 20),
        (b'fdsa', b'fdsa'),
    ),
)
def test_simple_binary_serialization(value, expected):
    sedes = Binary()
    assert sedes.serialize(value) == expected


@pytest.mark.parametrize(
    'value',
    ([], 5, str, '', 'arst'),
)
def test_binary_unserializable_values(value):
    sedes = Binary()
    with pytest.raises(SerializationError):
        sedes.serialize(value)


@pytest.mark.parametrize(
    'value,expected',
    (
        (b'asdfg', b'asdfg'),
        (b'\x00\x01\x02\x03\x04', b'\x00\x01\x02\x03\x04'),
        (b'ababa', b'ababa'),
    )
)
def test_binary_fixed_length_serialization(value, expected):
    sedes = Binary.fixed_length(5)
    assert sedes.serialize(value) == expected


def test_binary_fixed_lenght_of_zero():
    sedes = Binary.fixed_length(0)
    assert sedes.serialize(b'') == b''

    with pytest.raises(SerializationError):
        sedes.serialize(b'a')
    with pytest.raises(SerializationError):
        sedes.serialize(b'arst')


@pytest.mark.parametrize(
    'value',
    (b'asdf', b'asdfgh', b'', b'bababa'),
)
def test_binary_fixed_length_serialization_with_wrong_length(value):
    sedes = Binary.fixed_length(5)
    with pytest.raises(SerializationError):
        sedes.serialize(value)


@pytest.mark.parametrize(
    'value,expected',
    (
        (b'as', b'as'),
        (b'dfg', b'dfg'),
        (b'hjkl', b'hjkl'),
        (b'\x00\x01\x02', b'\x00\x01\x02'),
    ),
)
def test_binary_variable_length_serialization(value, expected):
    sedes = Binary(2, 4)
    assert sedes.serialize(value) == expected


@pytest.mark.parametrize(
    'value',
    (b'', b'a', b'abcde'),
)
def test_binary_variable_length_serialization_wrong_length(value):
    sedes = Binary(2, 4)
    with pytest.raises(SerializationError):
        sedes.serialize(value)


@pytest.mark.parametrize(
    'value,expected',
    (
        (b'abc', b'abc'),
        (b'abcd', b'abcd'),
        (b'x' * 132, b'x' * 132),
    ),
)
def test_binary_min_length_serialization(value, expected):
    sedes = Binary(min_length=3)
    assert sedes.serialize(value) == expected


@pytest.mark.parametrize(
    'value',
    (b'ab', b'', b'a', b'xy'),
)
def test_binary_min_length_serialization_wrong_length(value):
    sedes = Binary(min_length=3)
    with pytest.raises(SerializationError):
        sedes.serialize(value)


@pytest.mark.parametrize(
    'value,expected',
    (
        (b'', b''),
        (b'ab', b'ab'),
        (b'abc', b'abc'),
    )
)
def test_binary_max_length_serialization(value, expected):
    sedes = Binary(max_length=3)
    assert sedes.serialize(value) == expected


@pytest.mark.parametrize(
    'value',
    (b'abcd', b'vwxyz', b'a' * 32),
)
def test_binary_max_length_serialization_wrong_length(value):
    sedes = Binary(max_length=3)
    with pytest.raises(SerializationError):
        sedes.serialize(value)


@pytest.mark.parametrize(
    'value,expected',
    (
        (b'', b''),
        (b'abc', b'abc'),
        (b'abcd', b'abcd'),
        (b'abcde', b'abcde'),
    )
)
def test_binary_min_and_max_length_with_allow_empty(value, expected):
    sedes = Binary(min_length=3, max_length=5, allow_empty=True)
    assert sedes.serialize(value) == expected


@pytest.mark.parametrize(
    'value',
    (b'a', b'ab', b'abcdef', b'abcdefgh' * 10),
)
def test_binary_min_and_max_length_with_allow_empty_wrong_length(value):
    sedes = Binary(min_length=3, max_length=5, allow_empty=True)
    with pytest.raises(SerializationError):
        sedes.serialize(value)



import pytest
from rlp_cython import decode, DecodingError


invalid_rlp = (
    b'',
    b'\x00\xab',
    b'\x00\x00\xff',
    b'\x83dogcat',
    b'\x83do',
    b'\xc7\xc0\xc1\xc0\xc3\xc0\xc1\xc0\xff',
    b'\xc7\xc0\xc1\xc0\xc3\xc0\xc1'
    b'\x81\x02',
    b'\xb8\x00',
    b'\xb9\x00\x00',
    b'\xba\x00\x02\xff\xff',
    b'\x81\x54'
)


def test_invalid_rlp():
    for serial in invalid_rlp:
        with pytest.raises(DecodingError):
            decode(serial)



# -*- coding: UTF-8 -*-
import pytest
from rlp_cython import (
    SerializationError,
    DeserializationError,
    encode,
    decode,
)
from rlp_cython.sedes import Text

from hypothesis import (
    given,
    strategies as st,
)


@pytest.mark.parametrize(
    'value,expected',
    (
        ('', b''),
        ('asdf', b'asdf'),
        ('fdsa', b'fdsa'),
        ('�', b'\xef\xbf\xbd'),
        ("", b'\xc2\x80'),
        ("ࠀ", b'\xe0\xa0\x80'),
        ("𐀀", b'\xf0\x90\x80\x80'),
        ("�����", b'\xef\xbf\xbd\xef\xbf\xbd\xef\xbf\xbd\xef\xbf\xbd\xef\xbf\xbd'),
        ("������", b'\xef\xbf\xbd\xef\xbf\xbd\xef\xbf\xbd\xef\xbf\xbd\xef\xbf\xbd\xef\xbf\xbd'),
    ),
)
def test_simple_text_serialization(value, expected):
    sedes = Text()
    assert sedes.serialize(value) == expected


@pytest.mark.parametrize(
    'value',
    (
        b'',
        b'arst',
        1,
        True,
        None,
        1.0,
        bytearray(),
    ),
)
def test_unserializable_text_sedes_values(value):
    sedes = Text()
    with pytest.raises(SerializationError):
        sedes.serialize(value)


@pytest.mark.parametrize(
    'length,value,expected',
    (
        (0, '', b''),
        (1, 'a', b'a'),
        (1, 'a', b'a'),
        (1, '�', b'\xef\xbf\xbd'),
        (1, "", b'\xc2\x80'),
        (1, "ࠀ", b'\xe0\xa0\x80'),
        (1, "𐀀", b'\xf0\x90\x80\x80'),
    ),
)
def test_fixed_length_text_serialization(length, value, expected):
    sedes = Text.fixed_length(length)
    assert sedes.serialize(value) == expected


@pytest.mark.parametrize(
    'length,value',
    (
        (1, ''),
        (0, 'a'),
        (2, 'a'),
        (2, 'a'),
        (2, '�'),
        (4, '�'),  # actual binary length
        (2, ""),
        (4, ""),  # actual binary length
        (2, "ࠀ"),
        (4, "ࠀ"),  # actual binary length
        (2, "𐀀"),
        (4, "𐀀"),  # actual binary length
        (4, "�����"),
        (15, "�����"),  # actual binary length
        (5, "������"),
        (18, "������"),  # actual binary length
    ),
)
def test_fixed_length_text_serialization_with_wrong_length(length, value):
    sedes = Text.fixed_length(length)
    with pytest.raises(SerializationError):
        sedes.serialize(value)


@pytest.mark.parametrize(
    'min_length,max_length,value,expected',
    (
        (0, 4, '', b''),
        (0, 4, 'arst', b'arst'),
        (0, 4, 'arst', b'arst'),
        (0, 4, 'arst', b'arst'),
        (0, 1, '�', b'\xef\xbf\xbd'),
        (0, 1, "", b'\xc2\x80'),
        (0, 1, "ࠀ", b'\xe0\xa0\x80'),
        (0, 1, "𐀀", b'\xf0\x90\x80\x80'),
        (0, 5, "�����", b'\xef\xbf\xbd\xef\xbf\xbd\xef\xbf\xbd\xef\xbf\xbd\xef\xbf\xbd'),
        (0, 6, "������", b'\xef\xbf\xbd\xef\xbf\xbd\xef\xbf\xbd\xef\xbf\xbd\xef\xbf\xbd\xef\xbf\xbd'),  # noqa: E501
    ),
)
def test_min_max_length_text_serialization(min_length, max_length, value, expected):
    sedes = Text(min_length=min_length, max_length=max_length)
    assert sedes.serialize(value) == expected


@pytest.mark.parametrize(
    'min_length,max_length,value',
    (
        (1, 4, ''),
        (5, 6, ''),
        (1, 3, 'arst'),
        (5, 6, 'arst'),
        (2, 3, '�'),
        (4, 5, ""),
        (6, 7, "ࠀ"),
        (8, 9, "𐀀"),
        (6, 9, "�����"),
        (0, 4, "������"),
    ),
)
def test_min_max_length_text_serialization_wrong_length(min_length, max_length, value):
    sedes = Text(min_length=min_length, max_length=max_length)
    with pytest.raises(SerializationError):
        sedes.serialize(value)


@pytest.mark.parametrize(
    'serial,expected',
    (
        (b'', ''),
        (b'asdf', 'asdf'),
        (b'fdsa', 'fdsa'),
        (b'\xef\xbf\xbd', '�'),
        (b'\xc2\x80', ""),
        (b'\xe0\xa0\x80', "ࠀ"),
        (b'\xf0\x90\x80\x80', "𐀀"),
        (b'\xef\xbf\xbd\xef\xbf\xbd\xef\xbf\xbd\xef\xbf\xbd\xef\xbf\xbd', "�����"),
        (b'\xef\xbf\xbd\xef\xbf\xbd\xef\xbf\xbd\xef\xbf\xbd\xef\xbf\xbd\xef\xbf\xbd', "������"),
    ),
)
def test_deserialization_text_sedes(serial, expected):
    sedes = Text()
    assert sedes.deserialize(serial) == expected


def test_allow_empty_bypasses_length_checks():
    sedes = Text.fixed_length(1, allow_empty=True)
    assert sedes.serialize('') == b''

    with pytest.raises(SerializationError):
        sedes.serialize(b'12')


@given(value=st.text())
def test_round_trip_text_encoding_and_decoding(value):
    sedes = Text()
    encoded = encode(value, sedes=sedes)
    actual = decode(encoded, sedes=sedes)
    assert actual == value


def test_desirialization_of_text_encoding_failure():
    sedes = Text()
    with pytest.raises(DeserializationError):
        sedes.deserialize(b'\xff')



import pytest

from eth_utils import (
    decode_hex,
)

from rlp_cython.exceptions import DecodingError
from rlp_cython.codec import consume_length_prefix, consume_item
from rlp_cython import (
    decode,
    encode,
)


EMPTYLIST = encode([])


def compare_length(rlpdata, length):
    _, _typ, _len, _pos = consume_length_prefix(rlpdata, 0)
    assert _typ is list
    lenlist = 0
    if rlpdata == EMPTYLIST:
        return -1 if length > 0 else 1 if length < 0 else 0
    while 1:
        if lenlist > length:
            return 1
        _, _, _l, _p = consume_length_prefix(rlpdata, _pos)
        lenlist += 1
        if _l + _p >= len(rlpdata):
            break
        _pos = _l + _p
    return 0 if lenlist == length else -1


def test_compare_length():
    data = encode([1, 2, 3, 4, 5])
    assert compare_length(data, 100) == -1
    assert compare_length(data, 5) == 0
    assert compare_length(data, 1) == 1

    data = encode([])
    assert compare_length(data, 100) == -1
    assert compare_length(data, 0) == 0
    assert compare_length(data, -1) == 1


def test_favor_short_string_form():
    data = decode_hex('b8056d6f6f7365')
    with pytest.raises(DecodingError):
        decode(data)

    data = decode_hex('856d6f6f7365')
    assert decode(data) == b'moose'


def test_consume_item():
    obj = [b'f', b'bar', b'a' * 100, 105, [b'nested', b'list']]
    rlp = encode(obj)
    item, per_item_rlp, end = consume_item(rlp, 0)
    assert per_item_rlp == [
        (b'\xf8y' b'f' + b'\x83bar' + b'\xb8d' + b'a' * 100 + b'i' +
         b'\xcc\x86nested\x84list'),
        [b'f'],
        [b'\x83bar'],
        [b'\xb8d' + b'a' * 100],
        [b'i'],
        [b'\xcc\x86nested\x84list', [b'\x86nested'], [b'\x84list']]
    ]
    assert end == 123
    assert per_item_rlp[0] == rlp



class RLPException(Exception):
    """Base class for exceptions raised by this package."""
    pass


class EncodingError(RLPException):
    """Exception raised if encoding fails.

    :ivar obj: the object that could not be encoded
    """

    def __init__(self, message, obj):
        super(EncodingError, self).__init__(message)
        self.obj = obj


class DecodingError(RLPException):
    """Exception raised if decoding fails.

    :ivar rlp: the RLP string that could not be decoded
    """

    def __init__(self, message, rlp):
        super(DecodingError, self).__init__(message)
        self.rlp = rlp


class SerializationError(RLPException):
    """Exception raised if serialization fails.

    :ivar obj: the object that could not be serialized
    """

    def __init__(self, message, obj):
        super(SerializationError, self).__init__(message)
        self.obj = obj


class ListSerializationError(SerializationError):
    """Exception raised if serialization by a :class:`sedes.List` fails.

    :ivar element_exception: the exception that occurred during the serialization of one of the
                             elements, or `None` if the error is unrelated to a specific element
    :ivar index: the index in the list that produced the error or `None` if the error is unrelated
                 to a specific element
    """

    def __init__(self, message=None, obj=None, element_exception=None, index=None):
        if message is None:
            assert index is not None
            assert element_exception is not None
            message = ('Serialization failed because of element at index {} '
                       '("{}")'.format(index, str(element_exception)))
        super(ListSerializationError, self).__init__(message, obj)
        self.index = index
        self.element_exception = element_exception


class ObjectSerializationError(SerializationError):
    """Exception raised if serialization of a :class:`sedes.Serializable` object fails.

    :ivar sedes: the :class:`sedes.Serializable` that failed
    :ivar list_exception: exception raised by the underlying list sedes, or `None` if no such
                          exception has been raised
    :ivar field: name of the field of the object that produced the error, or `None` if no field
                 responsible for the error
    """

    def __init__(self, message=None, obj=None, sedes=None, list_exception=None):
        if message is None:
            assert list_exception is not None
            if list_exception.element_exception is None:
                field = None
                message = ('Serialization failed because of underlying list '
                           '("{}")'.format(str(list_exception)))
            else:
                assert sedes is not None
                field = sedes._meta.field_names[list_exception.index]
                message = ('Serialization failed because of field {} '
                           '("{}")'.format(field, str(list_exception.element_exception)))
        else:
            field = None
        super(ObjectSerializationError, self).__init__(message, obj)
        self.field = field
        self.list_exception = list_exception


class DeserializationError(RLPException):
    """Exception raised if deserialization fails.

    :ivar serial: the decoded RLP string that could not be deserialized
    """

    def __init__(self, message, serial):
        super(DeserializationError, self).__init__(message)
        self.serial = serial


class ListDeserializationError(DeserializationError):
    """Exception raised if deserialization by a :class:`sedes.List` fails.

    :ivar element_exception: the exception that occurred during the deserialization of one of the
                             elements, or `None` if the error is unrelated to a specific element
    :ivar index: the index in the list that produced the error or `None` if the error is unrelated
                 to a specific element
    """

    def __init__(self, message=None, serial=None, element_exception=None, index=None):
        if not message:
            assert index is not None
            assert element_exception is not None
            message = ('Deserialization failed because of element at index {} '
                       '("{}")'.format(index, str(element_exception)))
        super(ListDeserializationError, self).__init__(message, serial)
        self.index = index
        self.element_exception = element_exception


class ObjectDeserializationError(DeserializationError):
    """Exception raised if deserialization by a :class:`sedes.Serializable` fails.

    :ivar sedes: the :class:`sedes.Serializable` that failed
    :ivar list_exception: exception raised by the underlying list sedes, or `None` if no such
                          exception has been raised
    :ivar field: name of the field of the object that produced the error, or `None` if no field
                 responsible for the error
    """

    def __init__(self, message=None, serial=None, sedes=None, list_exception=None):
        if not message:
            assert list_exception is not None
            if list_exception.element_exception is None:
                field = None
                message = ('Deserialization failed because of underlying list '
                           '("{}")'.format(str(list_exception)))
            else:
                assert sedes is not None
                field = sedes._meta.field_names[list_exception.index]
                message = ('Deserialization failed because of field {} '
                           '("{}")'.format(field, str(list_exception.element_exception)))
        super(ObjectDeserializationError, self).__init__(message, serial)
        self.sedes = sedes
        self.list_exception = list_exception
        self.field = field



from rlp_cython.exceptions import SerializationError, DeserializationError
from rlp_cython.atomic import Atomic


class Binary(object):
    """A sedes object for binary data of certain length.

    :param min_length: the minimal length in bytes or `None` for no lower limit
    :param max_length: the maximal length in bytes or `None` for no upper limit
    :param allow_empty: if true, empty strings are considered valid even if
                        a minimum length is required otherwise
    """

    def __init__(self, min_length=None, max_length=None, allow_empty=False, validate = True):
        self.min_length = min_length or 0
        if max_length is None:
            self.max_length = float('inf')
        else:
            self.max_length = max_length
        self.allow_empty = allow_empty
        self.validate = validate

    @classmethod
    def fixed_length(cls, l, allow_empty=False):
        """Create a sedes for binary data with exactly `l` bytes."""
        return cls(l, l, allow_empty=allow_empty)

    @classmethod
    def is_valid_type(cls, obj):
        return isinstance(obj, (bytes, bytearray))

    def is_valid_length(self, l):
        return any((self.min_length <= l <= self.max_length,
                    self.allow_empty and l == 0))

    def get_sede_identifier(self):
        return 0

    # def serialize(self, obj):
    #     if self.validate and not Binary.is_valid_type(obj):
    #         raise SerializationError('Object is not a serializable ({})'.format(type(obj)), obj)
    #
    #     if self.validate and not self.is_valid_length(len(obj)):
    #         raise SerializationError('Object has invalid length', obj)
    #
    #     return obj
    #
    # def deserialize(self, serial):
    #     if self.validate and not isinstance(serial, Atomic):
    #         m = 'Objects of type {} cannot be deserialized'
    #         raise DeserializationError(m.format(type(serial).__name__), serial)
    #
    #     if not self.validate or self.is_valid_length(len(serial)):
    #         return serial
    #     else:
    #         raise DeserializationError('{} has invalid length'.format(type(serial)), serial)


    def serialize(self, obj):
        #Removed fixed length for now
        return obj

    def deserialize(self, serial, to_list = False):
        return serial


binary = Binary()



from collections import Sequence

import pytest

import rlp_cython as rlp
from rlp_cython import DeserializationError
from rlp_cython.sedes import big_endian_int, CountableList


def evaluate(lazy_list):
    if isinstance(lazy_list, rlp.lazy.LazyList):
        return tuple(evaluate(e) for e in lazy_list)
    else:
        return lazy_list


def test_empty_list():
    def dec():
        return rlp.decode_lazy(rlp.encode([]))

    assert isinstance(dec(), Sequence)
    with pytest.raises(IndexError):
        dec()[0]
    with pytest.raises(IndexError):
        dec()[1]
    assert len(dec()) == 0
    assert evaluate(dec()) == ()


@pytest.mark.parametrize(
    'value',
    (b'', b'asdf', b'a' * 56, b'b' * 123),
)
def test_string(value):
    def dec():
        return rlp.decode_lazy(rlp.encode(value))

    assert isinstance(dec(), bytes)
    assert len(dec()) == len(value)
    assert dec() == value
    assert rlp.peek(rlp.encode(value), []) == value
    with pytest.raises(IndexError):
        rlp.peek(rlp.encode(value), 0)
    with pytest.raises(IndexError):
        rlp.peek(rlp.encode(value), [0])


def test_list_getitem():
    l = rlp.decode_lazy(rlp.encode([1, 2, 3]), big_endian_int)
    assert isinstance(l, rlp.lazy.LazyList)
    assert l[0] == 1
    assert l[1] == 2
    assert l[2] == 3
    assert l[-1] == 3
    assert l[-2] == 2
    assert l[-3] == 1
    assert l[0:3] == [1, 2, 3]
    assert l[0:2] == [1, 2]
    assert l[0:1] == [1]
    assert l[1:2] == [2]
    assert l[1:] == [2, 3]
    assert l[1:-1] == [2]
    assert l[-2:] == [2, 3]
    assert l[:2] == [1, 2]


def test_nested_list():
    l = ((), (b'a'), (b'b', b'c', b'd'))

    def dec():
        return rlp.decode_lazy(rlp.encode(l))

    assert isinstance(dec(), Sequence)
    assert len(dec()) == len(l)
    assert evaluate(dec()) == l
    with pytest.raises(IndexError):
        dec()[0][0]
    with pytest.raises(IndexError):
        dec()[1][1]
    with pytest.raises(IndexError):
        dec()[2][3]
    with pytest.raises(IndexError):
        dec()[3]


@pytest.mark.parametrize(
    'value',
    (
        (),
        (1,),
        (3, 2, 1),
    ),
)
def test_evaluation_of_lazy_decode_with_simple_value_sedes(value):
    assert evaluate(rlp.decode_lazy(rlp.encode(value), big_endian_int)) == value


def test_evaluation_of_lazy_decode_with_list_sedes_and_invalid_value():
    sedes = CountableList(big_endian_int)
    value = [(), (1, 2), b'asdf', (3)]
    invalid_lazy = rlp.decode_lazy(rlp.encode(value), sedes)
    assert invalid_lazy[0] == value[0]
    assert invalid_lazy[1] == value[1]
    with pytest.raises(DeserializationError):
        invalid_lazy[2]


def test_peek():
    assert rlp.peek(rlp.encode(b''), []) == b''
    nested = rlp.encode([0, 1, [2, 3]])
    assert rlp.peek(nested, [2, 0], big_endian_int) == 2
    for index in [3, [3], [0, 0], [2, 2], [2, 1, 0]]:
        with pytest.raises(IndexError):
            rlp.peek(nested, index)
    assert rlp.peek(nested, 2, CountableList(big_endian_int)) == (2, 3)



from rlp_cython.exceptions import (
    DeserializationError,
    SerializationError,
)


class Boolean:
    """A sedes for booleans
    """
    def serialize(self, obj):
        if not isinstance(obj, bool):
            raise SerializationError('Can only serialize integers', obj)

        if obj is False:
            return b''
        elif obj is True:
            return b'\x01'
        else:
            raise Exception("Invariant: no other options for boolean values")

    def deserialize(self, serial, to_list = False):
        if serial == b'':
            return False
        elif serial == b'\x01':
            return True
        else:
            raise DeserializationError(
                'Invalid serialized boolean.  Must be either 0x01 or 0x00',
                serial
            )

    def get_sede_identifier(self):
        return 0

boolean = Boolean()



from eth_utils import (
    int_to_big_endian,
    big_endian_to_int,
)

from rlp_cython.exceptions import DeserializationError, SerializationError


class BigEndianInt(object):
    """A sedes for big endian integers.

    :param l: the size of the serialized representation in bytes or `None` to
              use the shortest possible one
    """

    def __init__(self, l=None):
        self.l = l

    def serialize(self, obj):
        if isinstance(obj, bool) or not isinstance(obj, int):
            raise SerializationError('Can only serialize integers', obj)
        if self.l is not None and obj >= 256**self.l:
            raise SerializationError('Integer too large (does not fit in {} '
                                     'bytes)'.format(self.l), obj)
        if obj < 0:
            raise SerializationError('Cannot serialize negative integers', obj)

        if obj == 0:
            s = b''
        else:
            s = int_to_big_endian(obj)

        if self.l is not None:
            return b'\x00' * max(0, self.l - len(s)) + s
        else:
            return s

    def get_sede_identifier(self):
        return 0

    def deserialize(self, serial, to_list = False):
        if self.l is not None and len(serial) != self.l:
            raise DeserializationError('Invalid serialization (wrong size)',
                                       serial)
        if self.l is None and len(serial) > 0 and serial[0:1] == b'\x00':
            raise DeserializationError('Invalid serialization (not minimal '
                                       'length)', serial)

        serial = serial or b'\x00'
        return big_endian_to_int(serial)



class FBigEndianInt(object):
    """A sedes for big endian integers.

    :param l: the size of the serialized representation in bytes or `None` to
              use the shortest possible one
    """

    def __init__(self, l=None):
        self.l = l


    def get_sede_identifier(self):
        return 1

    def serialize(self, obj):
        #doing int serialization in msgpack
        return obj


    def deserialize(self, serial, to_list = False):
        #doing int deserialization in msgpack
        return serial


big_endian_int = BigEndianInt()
f_big_endian_int = FBigEndianInt()



import collections

from eth_utils import (
    big_endian_to_int,
    int_to_big_endian,
    is_bytes,
)

from rlp_cython.atomic import (
    Atomic,
)
from rlp_cython.exceptions import EncodingError, DecodingError
from rlp_cython.sedes.binary import Binary as BinaryClass
from rlp_cython.sedes import big_endian_int, binary, boolean, text
from rlp_cython.sedes.lists import List, is_sedes, is_sequence
from rlp_cython.sedes.serializable import Serializable
from rlp_cython.utils import ALL_BYTES
from msgpack_rlp import packb, unpackb


def encode(obj, sedes=None, infer_serializer=True, cache=True):
    """Encode a Python object in RLP format.

    By default, the object is serialized in a suitable way first (using
    :func:`rlp.infer_sedes`) and then encoded. Serialization can be explicitly
    suppressed by setting `infer_serializer` to ``False`` and not passing an
    alternative as `sedes`.

    If `obj` has an attribute :attr:`_cached_rlp` (as, notably,
    :class:`rlp.Serializable`) and its value is not `None`, this value is
    returned bypassing serialization and encoding, unless `sedes` is given (as
    the cache is assumed to refer to the standard serialization which can be
    replaced by specifying `sedes`).

    If `obj` is a :class:`rlp.Serializable` and `cache` is true, the result of
    the encoding will be stored in :attr:`_cached_rlp` if it is empty.

    :param sedes: an object implementing a function ``serialize(obj)`` which will be used to
                  serialize ``obj`` before encoding, or ``None`` to use the infered one (if any)
    :param infer_serializer: if ``True`` an appropriate serializer will be selected using
                             :func:`rlp.infer_sedes` to serialize `obj` before encoding
    :param cache: cache the return value in `obj._cached_rlp` if possible
                  (default `True`)
    :returns: the RLP encoded item
    :raises: :exc:`rlp.EncodingError` in the rather unlikely case that the item is too big to
             encode (will not happen)
    :raises: :exc:`rlp.SerializationError` if the serialization fails
    """
    # if isinstance(obj, Serializable):
    #     cached_rlp = obj._cached_rlp
    #     if sedes is None and cached_rlp:
    #         return cached_rlp
    #     else:
    #         really_cache = (
    #             cache and
    #             sedes is None
    #         )
    # else:
    #     really_cache = False

    if sedes:
        item = sedes.serialize(obj)
    elif infer_serializer:
        item = infer_sedes(obj).serialize(obj)
    else:
        item = obj

    #result = encode_raw(item)
    # print(item)
    # print('test')
    # exit()
    result = packb(item)
    # if really_cache:
    #     obj._cached_rlp = result
    return result

def encode_raw(item):
    return packb(item)

# def encode_raw(item):
#     """RLP encode (a nested sequence of) :class:`Atomic`s."""
#     if isinstance(item, Atomic):
#         if len(item) == 1 and item[0] < 128:
#             return item
#         payload = item
#         prefix_offset = 128  # string
#     elif not isinstance(item, str) and isinstance(item, collections.Sequence):
#         payload = b''.join(encode_raw(x) for x in item)
#         prefix_offset = 192  # list
#     else:
#         msg = 'Cannot encode object of type {0}'.format(type(item).__name__)
#         raise EncodingError(msg, item)
#
#     try:
#         prefix = length_prefix(len(payload), prefix_offset)
#     except ValueError:
#         raise EncodingError('Item too big to encode', item)
#
#     return prefix + payload


LONG_LENGTH = 256**8


def length_prefix(length, offset):
    """Construct the prefix to lists or strings denoting their length.

    :param length: the length of the item in bytes
    :param offset: ``0x80`` when encoding raw bytes, ``0xc0`` when encoding a
                   list
    """
    if length < 56:
        return ALL_BYTES[offset + length]
    elif length < LONG_LENGTH:
        length_string = int_to_big_endian(length)
        return ALL_BYTES[offset + 56 - 1 + len(length_string)] + length_string
    else:
        raise ValueError('Length greater than 256**8')


SHORT_STRING = 128 + 56


def consume_length_prefix(rlp, start):
    """Read a length prefix from an RLP string.

    :param rlp: the rlp byte string to read from
    :param start: the position at which to start reading
    :returns: a tuple ``(prefix, type, length, end)``, where ``type`` is either ``str``
              or ``list`` depending on the type of the following payload,
              ``length`` is the length of the payload in bytes, and ``end`` is
              the position of the first payload byte in the rlp string
    """
    b0 = rlp[start]
    if b0 < 128:  # single byte
        return (b'', bytes, 1, start)
    elif b0 < SHORT_STRING:  # short string
        if b0 - 128 == 1 and rlp[start + 1] < 128:
            raise DecodingError('Encoded as short string although single byte was possible', rlp)
        return (rlp[start:start + 1], bytes, b0 - 128, start + 1)
    elif b0 < 192:  # long string
        ll = b0 - 183  # - (128 + 56 - 1)
        if rlp[start + 1:start + 2] == b'\x00':
            raise DecodingError('Length starts with zero bytes', rlp)
        len_prefix = rlp[start + 1:start + 1 + ll]
        l = big_endian_to_int(len_prefix)  # noqa: E741
        if l < 56:
            raise DecodingError('Long string prefix used for short string', rlp)
        return (rlp[start:start + 1] + len_prefix, bytes, l, start + 1 + ll)
    elif b0 < 192 + 56:  # short list
        return (rlp[start:start + 1], list, b0 - 192, start + 1)
    else:  # long list
        ll = b0 - 192 - 56 + 1
        if rlp[start + 1:start + 2] == b'\x00':
            raise DecodingError('Length starts with zero bytes', rlp)
        len_prefix = rlp[start + 1:start + 1 + ll]
        l = big_endian_to_int(len_prefix)  # noqa: E741
        if l < 56:
            raise DecodingError('Long list prefix used for short list', rlp)
        return (rlp[start:start + 1] + len_prefix, list, l, start + 1 + ll)


def consume_payload(rlp, prefix, start, type_, length):
    """Read the payload of an item from an RLP string.

    :param rlp: the rlp string to read from
    :param type_: the type of the payload (``bytes`` or ``list``)
    :param start: the position at which to start reading
    :param length: the length of the payload in bytes
    :returns: a tuple ``(item, per_item_rlp, end)``, where ``item`` is
              the read item, per_item_rlp is a list containing the RLP
              encoding of each item and ``end`` is the position of the
              first unprocessed byte
    """
    if type_ is bytes:
        item = rlp[start: start + length]
        return (item, [prefix + item], start + length)
    elif type_ is list:
        items = []
        per_item_rlp = []
        list_rlp = prefix
        next_item_start = start
        end = next_item_start + length
        while next_item_start < end:
            p, t, l, s = consume_length_prefix(rlp, next_item_start)
            item, item_rlp, next_item_start = consume_payload(rlp, p, s, t, l)
            per_item_rlp.append(item_rlp)
            # When the item returned above is a single element, item_rlp will also contain a
            # single element, but when it's a list, the first element will be the RLP of the
            # whole List, which is what we want here.
            list_rlp += item_rlp[0]
            items.append(item)
        per_item_rlp.insert(0, list_rlp)
        if next_item_start > end:
            raise DecodingError('List length prefix announced a too small '
                                'length', rlp)
        return (items, per_item_rlp, next_item_start)
    else:
        raise TypeError('Type must be either list or bytes')


def consume_item(rlp, start):
    """Read an item from an RLP string.

    :param rlp: the rlp string to read from
    :param start: the position at which to start reading
    :returns: a tuple ``(item, end)`` where ``item`` is the read item and
              ``end`` is the position of the first unprocessed byte
    """
    p, t, l, s = consume_length_prefix(rlp, start)
    return consume_payload(rlp, p, s, t, l)


def decode(rlp, sedes=None, strict=True, recursive_cache=False, use_list = False, **kwargs):
    """Decode an RLP encoded object.

    If the deserialized result `obj` has an attribute :attr:`_cached_rlp` (e.g. if `sedes` is a
    subclass of :class:`rlp.Serializable`) it will be set to `rlp`, which will improve performance
    on subsequent :func:`rlp.encode` calls. Bear in mind however that `obj` needs to make sure that
    this value is updated whenever one of its fields changes or prevent such changes entirely
    (:class:`rlp.sedes.Serializable` does the latter).

    :param sedes: an object implementing a function ``deserialize(code)`` which will be applied
                  after decoding, or ``None`` if no deserialization should be performed
    :param \*\*kwargs: additional keyword arguments that will be passed to the deserializer
    :param strict: if false inputs that are longer than necessary don't cause an exception
    :returns: the decoded and maybe deserialized Python object
    :raises: :exc:`rlp.DecodingError` if the input string does not end after the root item and
             `strict` is true
    :raises: :exc:`rlp.DeserializationError` if the deserialization fails
    """
    if sedes:
        try:
            fast_sedes = sedes.get_sede_identifier()
        except AttributeError:
            # Default an identifier of 0, which is binary. This can be used in the case where the python sede converts to binary.
            fast_sedes = 0
        item = unpackb(rlp, sedes=fast_sedes, use_list = use_list)
    else:
        item = unpackb(rlp, use_list = True) #For some fucking reason, if there are no sedes we the specifications say to return a list
    # if not is_bytes(rlp):
    #     raise DecodingError('Can only decode RLP bytes, got type %s' % type(rlp).__name__, rlp)
    # try:
    #    item, per_item_rlp, end = consume_item(rlp, 0)
    # except IndexError:
    #     raise DecodingError('RLP string too short', rlp)
    #this is checked withing msgpack
    # if end != len(rlp) and strict:
    #     msg = 'RLP string ends with {} superfluous bytes'.format(len(rlp) - end)
    #     raise DecodingError(msg, rlp)
    if sedes:
        obj = sedes.deserialize(item, to_list = use_list, **kwargs)
        # if is_sequence(obj) or hasattr(obj, '_cached_rlp'):
        #     _apply_rlp_cache(obj, per_item_rlp, recursive_cache)
        return obj
    else:
        return item


def _apply_rlp_cache(obj, split_rlp, recursive):
    item_rlp = split_rlp.pop(0)
    if isinstance(obj, (int, bool, str, bytes, bytearray)):
        return
    elif hasattr(obj, '_cached_rlp'):
        obj._cached_rlp = item_rlp
    if not recursive:
        return
    for sub in obj:
        if isinstance(sub, (int, bool, str, bytes, bytearray)):
            split_rlp.pop(0)
        else:
            sub_rlp = split_rlp.pop(0)
            _apply_rlp_cache(sub, sub_rlp, recursive)


def infer_sedes(obj):
    """Try to find a sedes objects suitable for a given Python object.

    The sedes objects considered are `obj`'s class, `big_endian_int` and
    `binary`. If `obj` is a sequence, a :class:`rlp.sedes.List` will be
    constructed recursively.

    :param obj: the python object for which to find a sedes object
    :raises: :exc:`TypeError` if no appropriate sedes could be found
    """
    if is_sedes(obj.__class__):
        return obj.__class__
    elif not isinstance(obj, bool) and isinstance(obj, int) and obj >= 0:
        return big_endian_int
    elif BinaryClass.is_valid_type(obj):
        return binary
    elif not isinstance(obj, str) and isinstance(obj, collections.Sequence):
        return List(map(infer_sedes, obj))

    msg = 'Did not find sedes handling type {}'.format(type(obj).__name__)
    raise TypeError(msg)



from __future__ import unicode_literals

import pytest
from rlp_cython import SerializationError, DeserializationError
from rlp_cython import infer_sedes
from rlp_cython.sedes import (
    big_endian_int,
    binary,
    boolean,
    text,
    CountableList,
    List,
)


@pytest.mark.parametrize(
    'value,expected',
    (
        (5, big_endian_int),
        (0, big_endian_int),
        (-1, None),
        (True, boolean),
        (False, boolean),
        (None, None),
        (b'', binary),
        (b'asdf', binary),
        (b'\xe4\xf6\xfc\xea\xe2\xfb', binary),
        ('', text),
        ('asdf', text),
        ('\xe4\xf6\xfc\xea\xe2\xfb', text),
        ('你好世界', text),
        ('\u4f60\u597d\u4e16\u754c', text),
        ([], List()),
        ([1, 2, 3], List((big_endian_int,) * 3)),
        ([[], b'asdf'], List(([], binary))),
        ([1, 'asdf'], List((big_endian_int, text))),
    ),
)
def test_inference(value, expected):
    if expected is not None:
        inferred = infer_sedes(value)
        assert inferred == expected
        expected.serialize(value)
    else:
        with pytest.raises(TypeError):
            infer_sedes(value)


def test_list_sedes():
    l1 = List()
    l2 = List((big_endian_int, big_endian_int))
    l3 = List((l1, l2, [[[]]]))

    l1.serialize([])
    l2.serialize((2, 3))
    l3.serialize([[], [5, 6], [[[]]]])

    with pytest.raises(SerializationError):
        l1.serialize([[]])
    with pytest.raises(SerializationError):
        l1.serialize([5])

    for d in ([], [1, 2, 3], [1, [2, 3], 4]):
        with pytest.raises(SerializationError):
            l2.serialize(d)
    for d in ([], [[], [], [[[]]]], [[], [5, 6], [[]]]):
        with pytest.raises(SerializationError):
            l3.serialize(d)

    c = CountableList(big_endian_int)
    assert l1.deserialize(c.serialize([])) == ()
    for s in (c.serialize(l) for l in [[1], [1, 2, 3], range(30), (4, 3)]):
        with pytest.raises(DeserializationError):
            l1.deserialize(s)

    valid = [(1, 2), (3, 4), (9, 8)]
    for s, v in ((c.serialize(v), v) for v in valid):
        assert l2.deserialize(s) == v
    invalid = [[], [1], [1, 2, 3]]
    for s in (c.serialize(i) for i in invalid):
        with pytest.raises(DeserializationError):
            l2.deserialize(s)



from collections import Iterable, Sequence

from .codec import consume_length_prefix, consume_payload
from .exceptions import DecodingError
from .atomic import Atomic


def decode_lazy(rlp, sedes=None, **sedes_kwargs):
    """Decode an RLP encoded object in a lazy fashion.

    If the encoded object is a bytestring, this function acts similar to
    :func:`rlp.decode`. If it is a list however, a :class:`LazyList` is
    returned instead. This object will decode the string lazily, avoiding
    both horizontal and vertical traversing as much as possible.

    The way `sedes` is applied depends on the decoded object: If it is a string
    `sedes` deserializes it as a whole; if it is a list, each element is
    deserialized individually. In both cases, `sedes_kwargs` are passed on.
    Note that, if a deserializer is used, only "horizontal" but not
    "vertical lazyness" can be preserved.

    :param rlp: the RLP string to decode
    :param sedes: an object implementing a method ``deserialize(code)`` which
                  is used as described above, or ``None`` if no
                  deserialization should be performed
    :param \*\*sedes_kwargs: additional keyword arguments that will be passed
                             to the deserializers
    :returns: either the already decoded and deserialized object (if encoded as
              a string) or an instance of :class:`rlp.LazyList`
    """
    item, end = consume_item_lazy(rlp, 0)
    if end != len(rlp):
        raise DecodingError('RLP length prefix announced wrong length', rlp)
    if isinstance(item, LazyList):
        item.sedes = sedes
        item.sedes_kwargs = sedes_kwargs
        return item
    elif sedes:
        return sedes.deserialize(item, **sedes_kwargs)
    else:
        return item


def consume_item_lazy(rlp, start):
    """Read an item from an RLP string lazily.

    If the length prefix announces a string, the string is read; if it
    announces a list, a :class:`LazyList` is created.

    :param rlp: the rlp string to read from
    :param start: the position at which to start reading
    :returns: a tuple ``(item, end)`` where ``item`` is the read string or a
              :class:`LazyList` and ``end`` is the position of the first
              unprocessed byte.
    """
    p, t, l, s = consume_length_prefix(rlp, start)
    if t is bytes:
        item, _, end = consume_payload(rlp, p, s, bytes, l)
        return item, end
    else:
        assert t is list
        return LazyList(rlp, s, s + l), s + l


class LazyList(Sequence):
    """A RLP encoded list which decodes itself when necessary.

    Both indexing with positive indices and iterating are supported.
    Getting the length with :func:`len` is possible as well but requires full
    horizontal encoding.

    :param rlp: the rlp string in which the list is encoded
    :param start: the position of the first payload byte of the encoded list
    :param end: the position of the last payload byte of the encoded list
    :param sedes: a sedes object which deserializes each element of the list,
                  or ``None`` for no deserialization
    :param \*\*sedes_kwargs: keyword arguments which will be passed on to the
                             deserializer
    """

    def __init__(self, rlp, start, end, sedes=None, **sedes_kwargs):
        self.rlp = rlp
        self.start = start
        self.end = end
        self.index = start
        self._elements = []
        self._len = None
        self.sedes = sedes
        self.sedes_kwargs = sedes_kwargs

    def next(self):
        if self.index == self.end:
            self._len = len(self._elements)
            raise StopIteration
        assert self.index < self.end
        item, end = consume_item_lazy(self.rlp, self.index)
        self.index = end
        if self.sedes:
            item = self.sedes.deserialize(item, **self.sedes_kwargs)
        self._elements.append(item)
        return item

    def __getitem__(self, i):
        if isinstance(i, slice):
            if i.step is not None:
                raise TypeError("Step not supported")
            start = i.start
            stop = i.stop
        else:
            start = i
            stop = i + 1

        if stop is None:
            stop = self.end - 1

        try:
            while len(self._elements) < stop:
                self.next()
        except StopIteration:
            assert self.index == self.end
            raise IndexError('Index %s out of range' % i)

        if isinstance(i, slice):
            return self._elements[start:stop]
        else:
            return self._elements[start]

    def __len__(self):
        if not self._len:
            try:
                while True:
                    self.next()
            except StopIteration:
                self._len = len(self._elements)
        return self._len


def peek(rlp, index, sedes=None):
    """Get a specific element from an rlp encoded nested list.

    This function uses :func:`rlp.decode_lazy` and, thus, decodes only the
    necessary parts of the string.

    Usage example::

        >>> import rlp
        >>> rlpdata = rlp.encode([1, 2, [3, [4, 5]]])
        >>> rlp.peek(rlpdata, 0, rlp.sedes.big_endian_int)
        1
        >>> rlp.peek(rlpdata, [2, 0], rlp.sedes.big_endian_int)
        3

    :param rlp: the rlp string
    :param index: the index of the element to peek at (can be a list for
                  nested data)
    :param sedes: a sedes used to deserialize the peeked at object, or `None`
                  if no deserialization should be performed
    :raises: :exc:`IndexError` if `index` is invalid (out of range or too many
             levels)
    """
    ll = decode_lazy(rlp)
    if not isinstance(index, Iterable):
        index = [index]
    for i in index:
        if isinstance(ll, Atomic):
            raise IndexError('Too many indices given')
        ll = ll[i]
    if sedes:
        return sedes.deserialize(ll)
    else:
        return ll



import pytest
import rlp_cython as rlp
from rlp_cython.sedes import big_endian_int
from rlp_cython.sedes.lists import CountableList
from rlp_cython import SerializationError, DeserializationError


def test_countable_list():
    l1 = CountableList(big_endian_int)
    serializable = [(), (1, 2), tuple(range(500))]
    for s in serializable:
        r = l1.serialize(s)
        assert l1.deserialize(r) == s
    not_serializable = ([1, 'asdf'], ['asdf'], [1, [2]], [[]])
    for n in not_serializable:
        with pytest.raises(SerializationError):
            l1.serialize(n)

    l2 = CountableList(CountableList(big_endian_int))
    serializable = ((), ((),), ((1, 2, 3), (4,)), ((5,), (6, 7, 8)), ((), (),
                    (9, 0)))
    for s in serializable:
        r = l2.serialize(s)
        assert l2.deserialize(r) == s
    not_serializable = ([[[]]], [1, 2], [1, ['asdf'], ['fdsa']])
    for n in not_serializable:
        with pytest.raises(SerializationError):
            l2.serialize(n)

    l3 = CountableList(big_endian_int, max_length=3)
    serializable = [(), (1,), (1, 2), (1, 2, 3)]
    for s in serializable:
        r = l3.serialize(s)
        assert r == l1.serialize(s)
        assert l3.deserialize(r) == s
    not_serializable = [(1, 2, 3, 4), (1, 2, 3, 4, 5, 6, 7), range(500)]
    for s in not_serializable:
        with pytest.raises(SerializationError):
            l3.serialize(s)
        r = l1.serialize(s)
        with pytest.raises(DeserializationError):
            l3.deserialize(r)
        ll = rlp.decode_lazy(rlp.encode(r))
        with pytest.raises(DeserializationError):
            l3.deserialize(ll)
        assert len(ll._elements) == 3 + 1  # failed early, did not consume fully



import pytest
from rlp_cython import encode, decode, SerializationError, DecodingError
from rlp_cython.sedes import raw


serializable = (
    b'',
    b'asdf',
    b'fds89032#$@%',
    b'',
    b'dfsa',
    [b'dfsa', b''],
    [],
    [b'fdsa', [b'dfs', [b'jfdkl']]],
)


def test_serializable():
    for s in serializable:
        raw.serialize(s)
        code = encode(s, raw)
        assert s == decode(code, raw)


@pytest.mark.parametrize(
    'rlp_data',
    (
        0,
        32,
        ['asdf', ['fdsa', [5]]],
        str
    ),
)
def test_invalid_serializations(rlp_data):
    with pytest.raises(SerializationError):
        raw.serialize(rlp_data)


@pytest.mark.parametrize(
    'rlp_data',
    (
        None,
        'asdf',
    ),
)
def test_invalid_deserializations(rlp_data):
    with pytest.raises(DecodingError):
        decode(rlp_data, raw)



"""
A sedes that does nothing. Thus, everything that can be directly encoded by RLP
is serializable. This sedes can be used as a placeholder when deserializing
larger structures.
"""
from collections import Sequence

from rlp_cython.exceptions import SerializationError
from rlp_cython.atomic import Atomic


def serializable(obj):
    if isinstance(obj, Atomic):
        return True
    elif not isinstance(obj, str) and isinstance(obj, Sequence):
        return all(map(serializable, obj))
    else:
        return False


def serialize(obj):
    if not serializable(obj):
        raise SerializationError('Can only serialize nested lists of strings',
                                 obj)
    return obj


def deserialize(serial):
    return serial



from . import raw  # noqa: F401
from .binary import Binary, binary  # noqa: F401
from .boolean import Boolean, boolean  # noqa: F401
from .big_endian_int import BigEndianInt, big_endian_int, f_big_endian_int  # noqa: F401
from .lists import CountableList, List, FCountableList, FList  # noqa: F401
from .serializable import Serializable  # noqa: F401
from .text import Text, text  # noqa: #401



# -*- coding: utf8 -*-
from rlp_cython import (
    encode,
    decode,
    decode_lazy,
)


def test_bytearray():
    e = encode(b'abc')
    expected = decode(e)
    actual = decode(bytearray(e))
    assert actual == expected


def test_bytearray_lazy():
    e = encode(b'abc')
    expected = decode(e)
    actual = decode_lazy(bytearray(e))
    assert expected == actual


def test_encoding_bytearray():
    s = b'abcdef'
    direct = encode(s)
    from_bytearray = encode(bytearray(s))
    assert direct == from_bytearray
    assert decode(direct) == s



"""
util to benchmark known usecase
"""
import random
import rlp_cython as rlp
from rlp_cython.sedes import big_endian_int, BigEndianInt, Binary, binary, CountableList
from rlp_cython.sedes.serializable import Serializable
import time

address = Binary.fixed_length(20, allow_empty=True)
int20 = BigEndianInt(20)
int32 = BigEndianInt(32)
int256 = BigEndianInt(256)
hash32 = Binary.fixed_length(32)
trie_root = Binary.fixed_length(32, allow_empty=True)


def zpad(x, l):
    return b'\x00' * max(0, l - len(x)) + x


class Transaction(Serializable):
    fields = [
        ('nonce', big_endian_int),
        ('gasprice', big_endian_int),
        ('startgas', big_endian_int),
        ('to', address),
        ('value', big_endian_int),
        ('data', binary),
        ('v', big_endian_int),
        ('r', big_endian_int),
        ('s', big_endian_int),
    ]

    def __init__(self, nonce, gasprice, startgas, to, value, data, v=0, r=0, s=0, **kwargs):
        super().__init__(nonce, gasprice, startgas, to, value, data, v, r, s, **kwargs)


class BlockHeader(Serializable):
    fields = [
        ('prevhash', hash32),
        ('uncles_hash', hash32),
        ('coinbase', address),
        ('state_root', trie_root),
        ('tx_list_root', trie_root),
        ('receipts_root', trie_root),
        ('bloom', int256),
        ('difficulty', big_endian_int),
        ('number', big_endian_int),
        ('gas_limit', big_endian_int),
        ('gas_used', big_endian_int),
        ('timestamp', big_endian_int),
        ('extra_data', binary),
        ('mixhash', binary),
        ('nonce', binary)
    ]


class Block(Serializable):
    fields = [
        ('header', BlockHeader),
        ('transaction_list', CountableList(Transaction)),
        ('uncles', CountableList(BlockHeader))
    ]

    def __init__(self, header, transaction_list=None, uncles=None, **kwargs):
        super().__init__(header, transaction_list or [], uncles or [], **kwargs)


def rand_bytes(num=32):
    return zpad(big_endian_int.serialize(random.getrandbits(num * 8)), num)


rand_bytes32 = rand_bytes


def rand_address():
    return rand_bytes(20)


def rand_bytes8():
    return rand_bytes(8)


def rand_bigint():
    return random.getrandbits(256)


def rand_int():
    return random.getrandbits(32)


rand_map = {hash32: rand_bytes32, trie_root: rand_bytes32, binary: rand_bytes32,
            address: rand_address, Binary: rand_bytes8, big_endian_int: rand_int,
            int256: rand_bigint}

assert Binary in rand_map


def mk_transaction():
    return Transaction(rand_int(), rand_int(), rand_int(), rand_address(), rand_int(),
                       rand_bytes32(), 27, rand_bigint(), rand_bigint())


rlp.decode(rlp.encode(mk_transaction()), Transaction)


def mk_block_header():
    return BlockHeader(*[rand_map[t]() for _, t in BlockHeader._meta.fields])


rlp.decode(rlp.encode(mk_block_header()), BlockHeader)


def mk_block(num_transactions=10, num_uncles=1):
    return Block(mk_block_header(),
                 [mk_transaction() for i in range(num_transactions)],
                 [mk_block_header() for i in range(num_uncles)])


rlp.decode(rlp.encode(mk_block()), Block)


def do_test_serialize(block, rounds=100):
    for i in range(rounds):
        x = rlp.encode(block, cache=False)
    return x


def do_test_deserialize(data, rounds=100, sedes=Block):
    for i in range(rounds):
        x = rlp.decode(data, sedes)
    return x


def main(rounds=10000):
    st = time.time()
    d = do_test_serialize(mk_block(), rounds)
    elapsed = time.time() - st
    print('Block serializations / sec: %.2f' % (rounds / elapsed))

    st = time.time()
    d = do_test_deserialize(d, rounds)
    elapsed = time.time() - st
    print('Block deserializations / sec: %.2f' % (rounds / elapsed))

    st = time.time()
    d = do_test_serialize(mk_transaction(), rounds)
    elapsed = time.time() - st
    print('TX serializations / sec: %.2f' % (rounds / elapsed))

    st = time.time()
    d = do_test_deserialize(d, rounds, sedes=Transaction)
    elapsed = time.time() - st
    print('TX deserializations / sec: %.2f' % (rounds / elapsed))


if __name__ == '__main__':
    main()
    """
    py2
    serializations / sec: 658.64
    deserializations / sec: 1331.62

    pypy2
    serializations / sec: 4628.81      : x7 speedup
    deserializations / sec: 4753.84    : x3.5 speedup
    """



import pytest

from rlp_cython import (
    SerializationError,
    DeserializationError,
)
from rlp_cython.sedes import Boolean


@pytest.mark.parametrize(
    'value,expected',
    (
        (True, b'\x01'),
        (False, b''),
    ),
)
def test_boolean_serialize_values(value, expected):
    sedes = Boolean()
    assert sedes.serialize(value) == expected


@pytest.mark.parametrize(
    'value',
    (
        None,
        1,
        0,
        'True',
        b'True',
    ),
)
def test_boolean_serialize_bad_values(value):
    sedes = Boolean()
    with pytest.raises(SerializationError):
        sedes.serialize(value)


@pytest.mark.parametrize(
    'value,expected',
    (
        (b'\x01', True),
        (b'', False),
    ),
)
def test_boolean_deserialization(value, expected):
    sedes = Boolean()
    assert sedes.deserialize(value) == expected


@pytest.mark.parametrize(
    'value',
    (
        b' ',
        b'\x02',
        b'\x00\x00',
        b'\x01\x00',
        b'\x00\x01',
        b'\x01\x01',
    ),
)
def test_boolean_deserialization_bad_value(value):
    sedes = Boolean()
    with pytest.raises(DeserializationError):
        sedes.deserialize(value)



import json

import pytest

from eth_utils import (
    decode_hex,
    encode_hex,
    add_0x_prefix,
)

import rlp_cython as rlp
from rlp_cython import encode, decode, decode_lazy, infer_sedes, DecodingError


def evaluate(ll):
    if isinstance(ll, rlp.lazy.LazyList):
        return [evaluate(e) for e in ll]
    else:
        return ll


def normalize_input(value):
    if isinstance(value, str):
        return value.encode('utf8')
    elif isinstance(value, list):
        return [normalize_input(v) for v in value]
    elif isinstance(value, int):
        return value
    else:
        raise ValueError("Unsupported type")


def compare_nested(got, expected):
    if isinstance(got, bytes):
        return got == expected
    try:
        zipped = zip(got, expected)
    except TypeError:
        return got == expected
    else:
        if len(list(zipped)) == len(got) == len(expected):
            return all(compare_nested(x, y) for x, y in zipped)
        else:
            return False


with open('tests/rlptest.json') as rlptest_file:
    test_data = json.load(rlptest_file)
    test_pieces = [
        (
            name,
            {
                'in': normalize_input(in_out['in']),
                'out': add_0x_prefix(in_out['out']),
            },
        )
        for name, in_out in test_data.items()
    ]


@pytest.mark.parametrize('name, in_out', test_pieces)
def test_encode(name, in_out):
    msg_format = 'Test {} failed (encoded {} to {} instead of {})'
    data = in_out['in']
    result = encode_hex(encode(data)).lower()
    expected = in_out['out'].lower()
    if result != expected:
        pytest.fail(msg_format.format(name, data, result, expected))


@pytest.mark.parametrize('name, in_out', test_pieces)
def test_decode(name, in_out):
    msg_format = 'Test {} failed (decoded {} to {} instead of {})'
    rlp_string = decode_hex(in_out['out'])
    decoded = decode(rlp_string)
    with pytest.raises(DecodingError):
        decode(rlp_string + b'\x00')
    assert decoded == decode(rlp_string + b'\x00', strict=False)

    assert decoded == evaluate(decode_lazy(rlp_string))
    expected = in_out['in']
    sedes = infer_sedes(expected)
    data = sedes.deserialize(decoded)
    assert compare_nested(data, decode(rlp_string, sedes))

    if not compare_nested(data, expected):
        pytest.fail(msg_format.format(name, rlp_string, decoded, expected))



#!/usr/bin/env python
# -*- coding: utf-8 -*-
from setuptools import (
    setup,
    find_packages,
)


extras_require = {
    'test': [
        "pytest==3.3.2",
        "tox>=2.9.1,<3",
        "hypothesis==3.56.5",
    ],
    'lint': [
        "flake8==3.4.1",
    ],
    'doc': [
        "Sphinx>=1.6.5,<2",
        "sphinx_rtd_theme>=0.1.9",
    ],
    'dev': [
        "bumpversion>=0.5.3,<1",
        "pytest-xdist",
        "pytest-watch>=4.1.0,<5",
        "wheel",
        "ipython",
        "twine",
    ],
}

dep = {
    'rlp': [
        "msgpack-rlp>=0.6",
        "eth-utils>=1.0.2,<2",
    ],
}

extras_require['dev'] = (
    extras_require['dev'] +
    extras_require['test'] +
    extras_require['lint'] +
    extras_require['doc'] +
    dep['rlp']
)

install_requires = dep['rlp']


setup(
    name='rlp-cython',
    # *IMPORTANT*: Don't manually change the version here. See README for more.
    version='2.1.2',
    description="A package for Recursive Length Prefix encoding and decoding",
    long_description_markdown_filename='README.md',
    author="Tommy Mckinnon",
    author_email='tommy@heliosprotocol.io',
    url='https://github.com/Helios-Protocol/py-rlp-cython',
    packages=find_packages(exclude=["tests", "tests.*"]),
    include_package_data=True,
    setup_requires=['setuptools-markdown'],
    install_requires=install_requires,
    extras_require=extras_require,
    license="MIT",
    zip_safe=False,
    keywords='rlp cython helios ethereum',
    classifiers=[
        'Intended Audience :: Developers',
        'License :: OSI Approved :: MIT License',
        'Natural Language :: English',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: Implementation :: PyPy',
    ],
)


from rlp_cython.exceptions import SerializationError, DeserializationError
from rlp_cython.atomic import Atomic


class Text:
    """A sedes object for encoded text data of certain length.

    :param min_length: the minimal length in encoded characters or `None` for no lower limit
    :param max_length: the maximal length in encoded characters or `None` for no upper limit
    :param allow_empty: if true, empty strings are considered valid even if
                        a minimum length is required otherwise
    """

    def __init__(self, min_length=None, max_length=None, allow_empty=False, encoding='utf8'):
        self.min_length = min_length or 0
        if max_length is None:
            self.max_length = float('inf')
        else:
            self.max_length = max_length
        self.allow_empty = allow_empty
        self.encoding = encoding

    @classmethod
    def fixed_length(cls, l, allow_empty=False):
        """Create a sedes for text data with exactly `l` encoded characters."""
        return cls(l, l, allow_empty=allow_empty)

    @classmethod
    def is_valid_type(cls, obj):
        return isinstance(obj, str)

    def is_valid_length(self, l):
        return any((
            self.min_length <= l <= self.max_length,
            self.allow_empty and l == 0
        ))

    def serialize(self, obj):
        if not self.is_valid_type(obj):
            raise SerializationError('Object is not a serializable ({})'.format(type(obj)), obj)

        if not self.is_valid_length(len(obj)):
            raise SerializationError('Object has invalid length', obj)

        return obj.encode(self.encoding)

    def deserialize(self, serial, to_list = False):
        if not isinstance(serial, Atomic):
            m = 'Objects of type {} cannot be deserialized'
            raise DeserializationError(m.format(type(serial).__name__), serial)

        try:
            text_value = serial.decode(self.encoding)
        except UnicodeDecodeError as err:
            raise DeserializationError(str(err), serial)

        if self.is_valid_length(len(text_value)):
            return text_value
        else:
            raise DeserializationError('{} has invalid length'.format(type(serial)), serial)


text = Text()



import re

import pytest

from rlp_cython import SerializationError
from rlp_cython import infer_sedes, encode, decode
from rlp_cython.sedes import big_endian_int, binary, List
from rlp_cython.sedes.serializable import Serializable


class RLPType1(Serializable):
    fields = [
        ('field1', big_endian_int),
        ('field2', binary),
        ('field3', List((big_endian_int, binary)))
    ]


class RLPType2(Serializable):
    fields = [
        ('field2_1', RLPType1),
        ('field2_2', List((RLPType1, RLPType1))),
    ]


class RLPType3(Serializable):
    fields = [
        ('field1', big_endian_int),
        ('field2', big_endian_int),
        ('field3', big_endian_int),
    ]

    def __init__(self, field2, field1, field3, **kwargs):
        super().__init__(field1=field1, field2=field2, field3=field3, **kwargs)


class RLPType4(RLPType3):
    pass


_type_1_a = RLPType1(5, b'a', (0, b''))
_type_1_b = RLPType1(9, b'b', (2, b''))
_type_2 = RLPType2(_type_1_a.copy(), [_type_1_a.copy(), _type_1_b.copy()])


@pytest.fixture
def type_1_a():
    return _type_1_a.copy()


@pytest.fixture
def type_1_b():
    return _type_1_b.copy()


@pytest.fixture
def type_2():
    return _type_2.copy()


@pytest.fixture(params=[_type_1_a, _type_1_b, _type_2])
def rlp_obj(request):
    return request.param.copy()


@pytest.mark.parametrize(
    'rlptype,args,kwargs,exception_includes',
    (
        # missing fields args
        (RLPType1, [], {}, ['field1', 'field2', 'field3']),
        (RLPType1, [8], {}, ['field2', 'field3']),
        (RLPType1, [7, 8], {}, ['field3']),
        # missing fields kwargs
        (RLPType1, [], {'field1': 7}, ['field2', 'field3']),
        (RLPType1, [], {'field1': 7, 'field2': 8}, ['field3']),
        (RLPType1, [], {'field2': 7, 'field3': (1, b'')}, ['field1']),
        (RLPType1, [], {'field3': (1, b'')}, ['field1', 'field2']),
        (RLPType1, [], {'field2': 7}, ['field1', 'field3']),
        # missing fields args and kwargs
        (RLPType1, [7], {'field2': 8}, ['field3']),
        (RLPType1, [7], {'field3': (1, b'')}, ['field2']),
        # duplicate fields
        (RLPType1, [7], {'field1': 8}, ['field1']),
        (RLPType1, [7, 8], {'field1': 8, 'field2': 7}, ['field1', 'field2']),
    ),
)
def test_serializable_initialization_validation(rlptype, args, kwargs, exception_includes):
    for msg_part in exception_includes:
        with pytest.raises(TypeError, match=msg_part):
            rlptype(*args, **kwargs)


@pytest.mark.parametrize(
    'args,kwargs',
    (
        ([2, 1, 3], {}),
        ([2, 1], {'field3': 3}),
        ([2], {'field3': 3, 'field1': 1}),
        ([], {'field3': 3, 'field1': 1, 'field2': 2}),
    ),
)
def test_serializable_initialization_args_kwargs_mix(args, kwargs):
    obj = RLPType3(*args, **kwargs)

    assert obj.field1 == 1
    assert obj.field2 == 2
    assert obj.field3 == 3


@pytest.mark.parametrize(
    'lookup,expected',
    (
        (0, 5),
        (1, b'a'),
        (2, (0, b'')),
        (slice(0, 1), (5,)),
        (slice(0, 2), (5, b'a')),
        (slice(0, 3), (5, b'a', (0, b''))),
        (slice(1, 3), (b'a', (0, b''))),
        (slice(2, 3), ((0, b''),)),
        (slice(None, 3), (5, b'a', (0, b''))),
        (slice(None, 2), (5, b'a')),
        (slice(None, 1), (5,)),
        (slice(None, 0), tuple()),
        (slice(0, None), (5, b'a', (0, b''))),
        (slice(1, None), (b'a', (0, b''))),
        (slice(2, None), ((0, b''),)),
        (slice(2, None), ((0, b''),)),
        (slice(None, None), (5, b'a', (0, b''))),
        (slice(-1, None), ((0, b''),)),
        (slice(-2, None), (b'a', (0, b''),)),
        (slice(-3, None), (5, b'a', (0, b''),)),
    ),
)
def test_serializable_getitem_lookups(type_1_a, lookup, expected):
    actual = type_1_a[lookup]
    assert actual == expected


def test_serializable_subclass_retains_field_info_from_parent():
    obj = RLPType4(2, 1, 3)
    assert obj.field1 == 1
    assert obj.field2 == 2
    assert obj.field3 == 3


def test_deserialization_for_custom_init_method():
    type_3 = RLPType3(2, 1, 3)
    assert type_3.field1 == 1
    assert type_3.field2 == 2
    assert type_3.field3 == 3

    result = decode(encode(type_3), sedes=RLPType3)

    assert result.field1 == 1
    assert result.field2 == 2
    assert result.field3 == 3


def test_serializable_iterator():
    values = (5, b'a', (1, b'c'))
    obj = RLPType1(*values)
    assert tuple(obj) == values


def test_serializable_equality(type_1_a, type_1_b, type_2):
    # equality
    assert type_1_a == type_1_a
    assert type_1_a == RLPType1(*type_1_a)
    assert type_1_b == type_1_b
    assert type_1_b == RLPType1(*type_1_b)

    assert type_2 == type_2
    assert type_1_a != type_1_b
    assert type_1_b != type_2
    assert type_2 != type_1_a


def test_serializable_sedes_inference(type_1_a, type_1_b, type_2):
    assert infer_sedes(type_1_a) == RLPType1
    assert infer_sedes(type_1_b) == RLPType1
    assert infer_sedes(type_2) == RLPType2


def test_serializable_invalid_serialization_value(type_1_a, type_1_b, type_2):
    with pytest.raises(SerializationError):
        RLPType1.serialize(type_2)
    with pytest.raises(SerializationError):
        RLPType2.serialize(type_1_a)
    with pytest.raises(SerializationError):
        RLPType2.serialize(type_1_b)


def test_serializable_serialization(type_1_a, type_1_b, type_2):
    serial_1_a = RLPType1.serialize(type_1_a)
    serial_1_b = RLPType1.serialize(type_1_b)
    serial_2 = RLPType2.serialize(type_2)
    assert serial_1_a == [b'\x05', b'a', [b'', b'']]
    assert serial_1_b == [b'\x09', b'b', [b'\x02', b'']]
    assert serial_2 == [serial_1_a, [serial_1_a, serial_1_b]]


def test_serializable_deserialization(type_1_a, type_1_b, type_2):
    serial_1_a = RLPType1.serialize(type_1_a)
    serial_1_b = RLPType1.serialize(type_1_b)
    serial_2 = RLPType2.serialize(type_2)

    res_type_1_a = RLPType1.deserialize(serial_1_a)
    res_type_1_b = RLPType1.deserialize(serial_1_b)
    res_type_2 = RLPType2.deserialize(serial_2)

    assert res_type_1_a == type_1_a
    assert res_type_1_b == type_1_b
    assert res_type_2 == type_2


def test_serializable_field_immutability(type_1_a, type_1_b, type_2):
    with pytest.raises(AttributeError, match=r"can't set attribute"):
        type_1_a.field1 += 1
    assert type_1_a.field1 == 5

    with pytest.raises(AttributeError, match=r"can't set attribute"):
        type_1_a.field2 = b'x'
    assert type_1_a.field2 == b'a'

    with pytest.raises(AttributeError, match=r"can't set attribute"):
        type_2.field2_1.field1 += 1
    assert type_2.field2_1.field1 == 5

    with pytest.raises(TypeError, match=r"'tuple' object does not support item assignment"):
        type_2.field2_2[1] = type_1_a
    assert type_2.field2_2[1] == type_1_b


def test_serializable_encoding_rlp_caching(rlp_obj):
    assert rlp_obj._cached_rlp is None

    # obj should start out without a cache
    rlp_code = encode(rlp_obj, cache=False)
    assert rlp_obj._cached_rlp is None

    # cache should be populated now.
    assert encode(rlp_obj, cache=True) == rlp_code
    assert rlp_obj._cached_rlp == rlp_code

    # cache should still be populated and encoding should used cached_rlp value
    rlp_obj._cached_rlp = b'test-uses-cache'
    assert encode(rlp_obj, cache=True) == b'test-uses-cache'

    obj_decoded = decode(rlp_code, sedes=rlp_obj.__class__)
    assert obj_decoded == rlp_obj
    assert obj_decoded._cached_rlp == rlp_code


def test_list_of_serializable_decoding_rlp_caching(rlp_obj):
    rlp_obj_code = encode(rlp_obj, cache=False)
    L = [rlp_obj, rlp_obj]
    list_code = encode(L, cache=False)

    L2 = decode(list_code, sedes=List((type(rlp_obj), type(rlp_obj))), recursive_cache=True)
    assert L2[0]._cached_rlp == rlp_obj_code
    assert L2[1]._cached_rlp == rlp_obj_code


def test_serializable_basic_copy(type_1_a):
    n_type_1_a = type_1_a.copy()
    assert n_type_1_a == type_1_a
    assert n_type_1_a is not type_1_a


def test_serializable_copy_with_nested_serializables(type_2):
    n_type_2 = type_2.copy()
    assert n_type_2 == type_2
    assert n_type_2 is not type_2

    assert n_type_2.field2_1 == type_2.field2_1
    assert n_type_2.field2_1 is not type_2.field2_1

    assert n_type_2.field2_2 == type_2.field2_2
    assert all(left == right for left, right in zip(n_type_2.field2_2, type_2.field2_2))
    assert all(left is not right for left, right in zip(n_type_2.field2_2, type_2.field2_2))


def test_serializable_build_changeset(type_1_a):
    with type_1_a.build_changeset() as changeset:
        # make changes to copy
        changeset.field1 = 1234
        changeset.field2 = b'arst'

        # check that the copy has the new field values
        assert changeset.field1 == 1234
        assert changeset.field2 == b'arst'

        n_type_1_a = changeset.commit()

    # check that the copy has the new field values
    assert n_type_1_a.field1 == 1234
    assert n_type_1_a.field2 == b'arst'

    # ensure the base object hasn't changed.
    assert type_1_a.field1 == 5
    assert type_1_a.field2 == b'a'


def test_serializable_build_changeset_changeset_gets_decomissioned(type_1_a):
    with type_1_a.build_changeset() as changeset:
        changeset.field1 = 54321
        n_type_1_a = changeset.commit()

    # ensure that we also can't update the changeset
    with pytest.raises(AttributeError):
        changeset.field1 = 12345
    # ensure that we also can't read values from the changeset
    with pytest.raises(AttributeError):
        changeset.field1

    assert n_type_1_a.field1 == 54321


def test_serializable_build_changeset_with_params(type_1_a):
    with type_1_a.build_changeset(1234) as changeset:
        assert changeset.field1 == 1234

        n_type_1_a = changeset.commit()
    assert n_type_1_a.field1 == 1234


def test_serializable_build_changeset_using_open_close_api(type_1_a):
    changeset = type_1_a.build_changeset()
    changeset.open()

    # make changes to copy
    changeset.field1 = 1234
    changeset.field2 = b'arst'

    # check that the copy has the new field values
    assert changeset.field1 == 1234
    assert changeset.field2 == b'arst'

    n_type_1_a = changeset.build_rlp()

    # check that the copy has the new field values
    assert n_type_1_a.field1 == 1234
    assert n_type_1_a.field2 == b'arst'

    # ensure the base object hasn't changed.
    assert type_1_a.field1 == 5
    assert type_1_a.field2 == b'a'

    # check we can still access the unclosed changeset
    assert changeset.field1 == 1234

    changeset.close()

    with pytest.raises(AttributeError):
        assert changeset.field1 == 1234


def test_serializable_with_duplicate_field_names_is_error():
    msg1 = "duplicated in the `fields` declaration: field_a"
    with pytest.raises(TypeError, match=msg1):
        class ParentA(Serializable):
            fields = (
                ('field_a', big_endian_int),
                ('field_c', big_endian_int),
                ('field_d', big_endian_int),
                ('field_a', big_endian_int),
            )

    msg2 = "duplicated in the `fields` declaration: field_a,field_c"
    with pytest.raises(TypeError, match=msg2):
        class ParentB(Serializable):
            fields = (
                ('field_a', big_endian_int),
                ('field_c', big_endian_int),
                ('field_d', big_endian_int),
                ('field_a', big_endian_int),
                ('field_c', big_endian_int),
            )


def test_serializable_inheritance_enforces_inclusion_of_parent_fields():
    class Parent(Serializable):
        fields = (
            ('field_a', big_endian_int),
            ('field_b', big_endian_int),
            ('field_c', big_endian_int),
            ('field_d', big_endian_int),
        )

    with pytest.raises(TypeError, match="field_a,field_c"):
        class Child(Parent):
            fields = (
                ('field_b', big_endian_int),
                ('field_d', big_endian_int),
            )


def test_serializable_single_inheritance_with_no_fields():
    class Parent(Serializable):
        fields = (
            ('field_a', big_endian_int),
            ('field_b', big_endian_int),
        )

    class Child(Parent):
        pass

    parent = Parent(1, 2)
    assert parent.field_a == 1
    assert parent.field_b == 2
    assert Parent.serialize(parent) == [b'\x01', b'\x02']

    child = Child(3, 4)
    assert child.field_a == 3
    assert child.field_b == 4
    assert Child.serialize(child) == [b'\x03', b'\x04']


def test_serializable_single_inheritance_with_fields():
    class Parent(Serializable):
        fields = (
            ('field_a', big_endian_int),
            ('field_b', big_endian_int),
        )

    class Child(Parent):
        fields = (
            ('field_a', big_endian_int),
            ('field_b', big_endian_int),
            ('field_c', big_endian_int),
        )

    parent = Parent(1, 2)
    assert parent.field_a == 1
    assert parent.field_b == 2
    assert Parent.serialize(parent) == [b'\x01', b'\x02']

    with pytest.raises(TypeError):
        # ensure that the fields don't somehow leak into the parent class.
        Parent(1, 2, 3)

    child = Child(3, 4, 5)
    assert child.field_a == 3
    assert child.field_b == 4
    assert child.field_c == 5
    assert Child.serialize(child) == [b'\x03', b'\x04', b'\x05']


def test_serializable_inheritance_with_sedes_overrides():
    class Parent(Serializable):
        fields = (
            ('field_a', big_endian_int),
            ('field_b', big_endian_int),
        )

    class Child(Parent):
        fields = (
            ('field_a', binary),
            ('field_b', binary),
            ('field_c', binary),
        )

    parent = Parent(1, 2)
    assert parent.field_a == 1
    assert parent.field_b == 2
    assert Parent.serialize(parent) == [b'\x01', b'\x02']

    child = Child(b'1', b'2', b'3')
    assert child.field_a == b'1'
    assert child.field_b == b'2'
    assert child.field_c == b'3'
    assert Child.serialize(child) == [b'1', b'2', b'3']


def test_serializable_multiple_inheritance_without_fields_declaration_is_error():
    class ParentA(Serializable):
        fields = (
            ('field_a', big_endian_int),
        )

    class ParentB(Serializable):
        fields = (
            ('field_b', big_endian_int),
        )

    with pytest.raises(TypeError, match="explicit `fields` declaration"):
        class Child(ParentA, ParentB):
            pass


def test_serializable_multiple_inheritance_allowed_with_explicit_fields():
    class ParentA(Serializable):
        fields = (
            ('field_a', big_endian_int),
        )

    class ParentB(Serializable):
        fields = (
            ('field_b', big_endian_int),
        )

    # with same fields
    class ChildA(ParentA, ParentB):
        fields = (
            ('field_a', big_endian_int),
            ('field_b', big_endian_int),
        )

    # with extra fields
    class ChildB(ParentA, ParentB):
        fields = (
            ('field_a', big_endian_int),
            ('field_b', big_endian_int),
            ('field_c', big_endian_int),
        )


def test_serializable_multiple_inheritance_requires_all_parent_fields():
    class ParentA(Serializable):
        fields = (
            ('field_a', big_endian_int),
        )

    class ParentB(Serializable):
        fields = (
            ('field_b', big_endian_int),
        )

    with pytest.raises(TypeError, match="The following fields are missing: field_b"):
        class ChildA(ParentA, ParentB):
            fields = (
                ('field_a', big_endian_int),
            )

    with pytest.raises(TypeError, match="The following fields are missing: field_a"):
        class ChildB(ParentA, ParentB):
            fields = (
                ('field_b', big_endian_int),
            )

    with pytest.raises(TypeError, match="The following fields are missing: field_a,field_b"):
        class ChildC(ParentA, ParentB):
            fields = (
                ('field_c', big_endian_int),
                ('field_d', big_endian_int),
            )


@pytest.mark.parametrize(
    'name',
    (
        '0_starts_with_digit',
        ' starts_with_space',
        '$starts_with_dollar',
        'has_dollar_$_inside',
        'has spaces',
    )
)
def test_serializable_field_names_must_be_valid_identifiers(name):
    msg = "not valid python identifiers: `{0}`".format(re.escape(name))
    with pytest.raises(TypeError, match=msg):
        class Klass(Serializable):
            fields = (
                (name, big_endian_int),
            )


def test_serializable_inheritance_from_base_with_no_fields():
    """
    ensure that we can create base classes of the base `Serializable` without
    declaring fields.
    """
    class ExtendedSerializable(Serializable):
        pass

    class FurtherExtendedSerializable(ExtendedSerializable):
        pass



import abc
import collections
import copy
import enum
import re

from eth_utils import (
    to_dict,
    to_set,
    to_tuple,
)

from rlp_cython.exceptions import (
    ListSerializationError,
    ObjectSerializationError,
    ListDeserializationError,
    ObjectDeserializationError,
)

from .lists import (
    List,
)


class MetaBase:
    fields = None
    field_names = None
    field_attrs = None
    sedes = None


def _get_duplicates(values):
    counts = collections.Counter(values)
    return tuple(
        item
        for item, num in counts.items()
        if num > 1
    )


def validate_args_and_kwargs(args, kwargs, arg_names, allow_missing=False):
    duplicate_arg_names = _get_duplicates(arg_names)
    if duplicate_arg_names:
        raise TypeError("Duplicate argument names: {0}".format(sorted(duplicate_arg_names)))

    needed_kwargs = arg_names[len(args):]
    used_kwargs = set(arg_names[:len(args)])

    duplicate_kwargs = used_kwargs.intersection(kwargs.keys())
    if duplicate_kwargs:
        raise TypeError("Duplicate kwargs: {0}".format(sorted(duplicate_kwargs)))

    unknown_kwargs = set(kwargs.keys()).difference(arg_names)
    if unknown_kwargs:
        raise TypeError("Unknown kwargs: {0}".format(sorted(unknown_kwargs)))

    missing_kwargs = set(needed_kwargs).difference(kwargs.keys())
    if not allow_missing and missing_kwargs:
        raise TypeError("Missing kwargs: {0}".format(sorted(missing_kwargs)))


@to_tuple
def merge_kwargs_to_args(args, kwargs, arg_names, allow_missing=False):
    validate_args_and_kwargs(args, kwargs, arg_names, allow_missing=allow_missing)

    needed_kwargs = arg_names[len(args):]

    yield from args
    for arg_name in needed_kwargs:
        yield kwargs[arg_name]


@to_dict
def merge_args_to_kwargs(args, kwargs, arg_names, allow_missing=False):
    validate_args_and_kwargs(args, kwargs, arg_names, allow_missing=allow_missing)

    yield from kwargs.items()
    for value, name in zip(args, arg_names):
        yield name, value


def _eq(left, right):
    """
    Equality comparison that allows for equality between tuple and list types
    with equivalent elements.
    """
    if isinstance(left, (tuple, list)) and isinstance(right, (tuple, list)):
        return len(left) == len(right) and all(_eq(*pair) for pair in zip(left, right))
    else:
        return left == right


class ChangesetState(enum.Enum):
    INITIALIZED = 'INITIALIZED'
    OPEN = 'OPEN'
    CLOSED = 'CLOSED'


class ChangesetField:
    field = None

    def __init__(self, field):
        self.field = field

    def __get__(self, instance, type=None):
        if instance is None:
            return self
        elif instance.__state__ is not ChangesetState.OPEN:
            raise AttributeError("Changeset is not active.  Attribute access not allowed")
        else:
            try:
                return instance.__diff__[self.field]
            except KeyError:
                return getattr(instance.__original__, self.field)

    def __set__(self, instance, value):
        if instance.__state__ is not ChangesetState.OPEN:
            raise AttributeError("Changeset is not active.  Attribute access not allowed")
        instance.__diff__[self.field] = value


class BaseChangeset:
    # reference to the original Serializable instance.
    __original__ = None
    # the state of this fieldset.  Initialized -> Open -> Closed
    __state__ = None
    # the field changes that have been made in this change
    __diff__ = None

    def __init__(self, obj, changes=None):
        self.__original__ = obj
        self.__state__ = ChangesetState.INITIALIZED
        self.__diff__ = changes or {}

    def commit(self):
        obj = self.build_rlp()
        self.close()
        return obj

    def build_rlp(self):
        if self.__state__ == ChangesetState.OPEN:
            field_kwargs = {
                name: self.__diff__.get(name, self.__original__[name])
                for name
                in self.__original__._meta.field_names
            }
            return type(self.__original__)(**field_kwargs)
        else:
            raise ValueError("Cannot open Changeset which is not in the OPEN state")

    def open(self):
        if self.__state__ == ChangesetState.INITIALIZED:
            self.__state__ = ChangesetState.OPEN
        else:
            raise ValueError("Cannot open Changeset which is not in the INITIALIZED state")

    def close(self):
        if self.__state__ == ChangesetState.OPEN:
            self.__state__ = ChangesetState.CLOSED
        else:
            raise ValueError("Cannot open Changeset which is not in the INITIALIZED state")

    def __enter__(self):
        if self.__state__ == ChangesetState.INITIALIZED:
            self.open()
            return self
        else:
            raise ValueError("Cannot open Changeset which is not in the INITIALIZED state")

    def __exit__(self, exc_type, exc_value, traceback):
        if self.__state__ == ChangesetState.OPEN:
            self.close()


def Changeset(obj, changes):
    namespace = {
        name: ChangesetField(name)
        for name
        in obj._meta.field_names
    }
    cls = type(
        "{0}Changeset".format(obj.__class__.__name__),
        (BaseChangeset,),
        namespace,
    )
    return cls(obj, changes)


class BaseSerializable(collections.Sequence):
    def __init__(self, *args, **kwargs):
        if kwargs:
            field_values = merge_kwargs_to_args(args, kwargs, self._meta.field_names)
        else:
            field_values = args

        if len(field_values) != len(self._meta.field_names):
            raise TypeError(
                'Argument count mismatch. expected {0} - got {1} - missing {2}'.format(
                    len(self._meta.field_names),
                    len(field_values),
                    ','.join(self._meta.field_names[len(field_values):]),
                )
            )

        for value, attr in zip(field_values, self._meta.field_attrs):
            setattr(self, attr, make_immutable(value))

    _cached_rlp = None

    def as_dict(self):
        return dict(
            (field, value)
            for field, value
            in zip(self._meta.field_names, self)
        )

    def __iter__(self):
        for attr in self._meta.field_attrs:
            yield getattr(self, attr)

    def __getitem__(self, idx):
        if isinstance(idx, int):
            attr = self._meta.field_attrs[idx]
            return getattr(self, attr)
        elif isinstance(idx, slice):
            field_slice = self._meta.field_attrs[idx]
            return tuple(getattr(self, field) for field in field_slice)
        elif isinstance(idx, str):
            return getattr(self, idx)
        else:
            raise IndexError("Unsupported type for __getitem__: {0}".format(type(idx)))

    def __len__(self):
        return len(self._meta.fields)

    def __eq__(self, other):
        return isinstance(other, Serializable) and hash(self) == hash(other)

    _hash_cache = None

    def __hash__(self):
        if self._hash_cache is None:
            self._hash_cache = hash(tuple(self))

        return self._hash_cache

    @classmethod
    def get_sede_identifier(cls):
        try:
            return cls._meta.sedes.get_sede_identifier()
        except AttributeError:
            return 0

    @classmethod
    def serialize(cls, obj):
        try:
            return cls._meta.sedes.serialize(obj)
        except ListSerializationError as e:
            raise ObjectSerializationError(obj=obj, sedes=cls, list_exception=e)

    @classmethod
    def deserialize(cls, serial, to_list = False, **extra_kwargs):

        try:
            values = cls._meta.sedes.deserialize(serial, to_list = to_list)
        except ListDeserializationError as e:
            raise ObjectDeserializationError(serial=serial, sedes=cls, list_exception=e)

        args_as_kwargs = merge_args_to_kwargs(values, {}, cls._meta.field_names)
        return cls(**args_as_kwargs, **extra_kwargs)

    def copy(self, *args, **kwargs):
        missing_overrides = set(
            self._meta.field_names
        ).difference(
            kwargs.keys()
        ).difference(
            self._meta.field_names[:len(args)]
        )
        unchanged_kwargs = {
            key: copy.deepcopy(value)
            for key, value
            in self.as_dict().items()
            if key in missing_overrides
        }
        combined_kwargs = dict(**unchanged_kwargs, **kwargs)
        all_kwargs = merge_args_to_kwargs(args, combined_kwargs, self._meta.field_names)
        return type(self)(**all_kwargs)

    def __copy__(self):
        return self.copy()

    def __deepcopy__(self, *args):
        return self.copy()

    _in_mutable_context = False

    def build_changeset(self, *args, **kwargs):
        args_as_kwargs = merge_args_to_kwargs(
            args,
            kwargs,
            self._meta.field_names,
            allow_missing=True,
        )
        return Changeset(self, changes=args_as_kwargs)


def make_immutable(value):
    if isinstance(value, list):
        return tuple(make_immutable(item) for item in value)
    else:
        return value


@to_tuple
def _mk_field_attrs(field_names, extra_namespace):
    namespace = set(field_names).union(extra_namespace)
    for field in field_names:
        while True:
            field = '_' + field
            if field not in namespace:
                namespace.add(field)
                yield field
                break


def _mk_field_property(field, attr):
    def field_fn_getter(self):
        return getattr(self, attr)

    def field_fn_setter(self, value):
        if not self._in_mutable_context:
            raise AttributeError("can't set attribute")
        setattr(self, attr, value)

    return property(field_fn_getter, field_fn_setter)


IDENTIFIER_REGEX = re.compile(r"^[^\d\W]\w*\Z", re.UNICODE)


def _is_valid_identifier(value):
    # Source: https://stackoverflow.com/questions/5474008/regular-expression-to-confirm-whether-a-string-is-a-valid-identifier-in-python  # noqa: E501
    if not isinstance(value, str):
        return False
    return bool(IDENTIFIER_REGEX.match(value))


@to_set
def _get_class_namespace(cls):
    if hasattr(cls, '__dict__'):
        yield from cls.__dict__.keys()
    if hasattr(cls, '__slots__'):
        yield from cls.__slots__


class SerializableBase(abc.ABCMeta):
    def __new__(cls, name, bases, attrs):
        super_new = super(SerializableBase, cls).__new__

        serializable_bases = tuple(b for b in bases if isinstance(b, SerializableBase))
        has_multiple_serializable_parents = len(serializable_bases) > 1
        is_serializable_subclass = any(serializable_bases)
        declares_fields = 'fields' in attrs

        if not is_serializable_subclass:
            # If this is the original creation of the `Serializable` class,
            # just create the class.
            return super_new(cls, name, bases, attrs)
        elif not declares_fields:
            if has_multiple_serializable_parents:
                raise TypeError(
                    "Cannot create subclass from multiple parent `Serializable` "
                    "classes without explicit `fields` declaration."
                )
            else:
                # This is just a vanilla subclass of a `Serializable` parent class.
                parent_serializable = serializable_bases[0]

                if hasattr(parent_serializable, '_meta'):
                    fields = parent_serializable._meta.fields
                else:
                    # This is a subclass of `Serializable` which has no
                    # `fields`, likely intended for further subclassing.
                    return super_new(cls, name, bases, attrs)
        else:
            # ensure that the `fields` property is a tuple of tuples to ensure
            # immutability.
            fields = tuple(tuple(field) for field in attrs.pop('fields'))

        # split the fields into names and sedes
        field_names, sedes = zip(*fields)

        # check that field names are unique
        duplicate_field_names = _get_duplicates(field_names)
        if duplicate_field_names:
            raise TypeError(
                "The following fields are duplicated in the `fields` "
                "declaration: "
                "{0}".format(",".join(sorted(duplicate_field_names)))
            )

        # check that field names are valid identifiers
        invalid_field_names = {
            field_name
            for field_name
            in field_names
            if not _is_valid_identifier(field_name)
        }
        if invalid_field_names:
            raise TypeError(
                "The following field names are not valid python identifiers: {0}".format(
                    ",".join("`{0}`".format(item) for item in sorted(invalid_field_names))
                )
            )

        # extract all of the fields from parent `Serializable` classes.
        parent_field_names = {
            field_name
            for base in serializable_bases if hasattr(base, '_meta')
            for field_name in base._meta.field_names
        }

        # check that all fields from parent serializable classes are
        # represented on this class.
        missing_fields = parent_field_names.difference(field_names)
        if missing_fields:
            raise TypeError(
                "Subclasses of `Serializable` **must** contain a full superset "
                "of the fields defined in their parent classes.  The following "
                "fields are missing: "
                "{0}".format(",".join(sorted(missing_fields)))
            )

        # the actual field values are stored in separate *private* attributes.
        # This computes attribute names that don't conflict with other
        # attributes already present on the class.
        reserved_namespace = set(attrs.keys()).union(
            attr
            for base in bases
            for parent_cls in base.__mro__
            for attr in _get_class_namespace(parent_cls)
        )
        field_attrs = _mk_field_attrs(field_names, reserved_namespace)

        # construct the Meta object to store field information for the class
        meta_namespace = {
            'fields': fields,
            'field_attrs': field_attrs,
            'field_names': field_names,
            'sedes': List(sedes),
        }

        meta_base = attrs.pop('_meta', MetaBase)
        meta = type(
            'Meta',
            (meta_base,),
            meta_namespace,
        )
        attrs['_meta'] = meta

        # construct `property` attributes for read only access to the fields.
        field_props = tuple(
            (field, _mk_field_property(field, attr))
            for field, attr
            in zip(meta.field_names, meta.field_attrs)
        )

        return super_new(
            cls,
            name,
            bases,
            dict(
                field_props +
                tuple(attrs.items())
            ),
        )


class Serializable(BaseSerializable, metaclass=SerializableBase):
    """
    The base class for serializable objects.
    """
    pass
